import contextlib
import logging
from typing import List
from typing import Optional

import pyarrow
import pyarrow.compute as pc
import pyarrow.parquet as pq
from google.protobuf import timestamp_pb2

from tecton_core.errors import TectonValidationError
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper as FeatureDefinition
from tecton_core.offline_store import TIME_PARTITION
from tecton_core.offline_store import partition_size_for_delta
from tecton_core.offline_store import timestamp_formats
from tecton_core.schema_validation import cast
from tecton_materialization.ray.delta import DeltaWriter
from tecton_materialization.ray.job_status import JobStatusClient
from tecton_materialization.ray.materialization_utils import get_delta_writer
from tecton_materialization.ray.materialization_utils import write_to_online_store
from tecton_proto.materialization.job_metadata__client_pb2 import TectonManagedStage
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams
from tecton_proto.offlinestore.delta import metadata__client_pb2 as metadata_pb2


logger = logging.getLogger(__name__)


def _write_to_offline_store(
    delta_writer: DeltaWriter, materialized_data: pyarrow.Table, ingest_path: str
) -> Optional[List[str]]:
    transaction_metadata = metadata_pb2.TectonDeltaMetadata(ingest_path=ingest_path)
    if delta_writer.transaction_exists(transaction_metadata):
        return None

    @delta_writer.transaction(transaction_metadata)
    def txn():
        return delta_writer.write(materialized_data)

    return txn()


def ingest_pushed_df(
    materialization_task_params: MaterializationTaskParams, fd: FeatureDefinition, job_status_client: JobStatusClient
):
    """
    Triggers materialization from the specified parquet file.

    # NOTE: The Rift version of this job is a bit different compared to Spark:
    - Spark does not require writing to the offline store first.
    - For the offline store, we're currently appending the rows. Spark overwrites existing rows and only inserts new ones.
    """
    ingest_task_info = materialization_task_params.ingest_task_info
    ingest_path = ingest_task_info.ingest_parameters.ingest_path

    if not fd.writes_to_offline_store:
        msg = f"Offline materialization is required for FeatureTables on Rift {fd.id} ({fd.name})"
        raise Exception(msg)
    if not fd.has_delta_offline_store:
        msg = f"Delta is required for FeatureTables {fd.id} ({fd.name})"
        raise Exception(msg)

    table_for_ingest = pq.read_table(ingest_path)

    timestamp_key = fd.timestamp_key
    assert timestamp_key is not None

    if timestamp_key not in table_for_ingest.column_names:
        msg = f"Timestamp column {timestamp_key} was not present in the ingested dataframe"
        raise TectonValidationError(msg)

    # Validate the schema and normalize types
    table_for_ingest = cast(table_for_ingest, fd.view_schema)

    # Add partition column. It needs to be present before sending the data over to the offline store writer.
    # TODO: switch to DuckDB - https://tecton.atlassian.net/browse/TEC-18995
    partition_size = partition_size_for_delta(fd).as_timedelta()
    ts_format = timestamp_formats(partition_size).python_format
    partition_column = pc.strftime(table_for_ingest[timestamp_key], format=ts_format)
    table_for_ingest = table_for_ingest.append_column(pyarrow.field(TIME_PARTITION, pyarrow.string()), partition_column)

    delta_writer = get_delta_writer(materialization_task_params)
    assert ingest_task_info.ingest_parameters.write_to_offline_feature_store, "must write to the offline feature store"
    logger.info(f"Ingesting to the OfflineStore FT: {fd.id}")
    offline_stage_monitor = job_status_client.create_stage_monitor(
        TectonManagedStage.StageType.OFFLINE_STORE,
        "Unload features to offline store",
    )
    with offline_stage_monitor():
        parts = _write_to_offline_store(delta_writer, table_for_ingest, ingest_path)

    if ingest_task_info.ingest_parameters.write_to_online_feature_store:
        max_timestamp = pc.max(table_for_ingest[timestamp_key]).as_py()
        raw_data_end_time_epoch = timestamp_pb2.Timestamp()
        raw_data_end_time_epoch.FromDatetime(max_timestamp)

        logger.info(f"Ingesting to the OnlineStore FT: {fd.id}")
        online_stage_monitor = job_status_client.create_stage_monitor(
            TectonManagedStage.StageType.ONLINE_STORE,
            "Unload features to online store",
        )
        with online_stage_monitor(), contextlib.ExitStack() as stack:
            if parts is None:
                parts = delta_writer.write(table_for_ingest)
                stack.callback(delta_writer.abort)
            for uri in parts:
                write_to_online_store(
                    materialization_task_params.online_store_writer_config,
                    materialization_task_params.feature_view,
                    raw_data_end_time_epoch,
                    fd,
                    uri,
                )



from __future__ import annotations

import typing
from abc import abstractmethod
from typing import Callable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import attrs
import pandas
import pyarrow

from tecton_core.query import node_interface
from tecton_core.query.pandas.sql import SqlExecutor
from tecton_core.secrets import SecretResolver


if typing.TYPE_CHECKING:
    import snowflake.snowpark

# SqlExecNodes are responsible for executing the sql string produced by the query node and using its SqlExecutor to
# output a pandas dataframe


@attrs.frozen
class SqlExecNode:
    columns: List[str]
    sql_string: str
    sql_executor: SqlExecutor

    @classmethod
    def from_sql_inputs(
        cls, query_node: node_interface.QueryNode, sql_executor: SqlExecutor, pretty_sql: bool = False
    ) -> "SqlExecNode":
        return cls(
            columns=query_node.columns, sql_string=query_node.to_sql(pretty_sql=pretty_sql), sql_executor=sql_executor
        )  # type: ignore

    def to_dataframe(self) -> pandas.DataFrame:
        df = self._to_dataframe()
        # TODO(TEC-15985): Because we do not refresh schemas on data sources, we can sometimes get different columns than what we have
        # cached.
        if {c.lower() for c in df.columns} != {c.lower() for c in self.columns}:
            pass
            # Because we do not refresh schemas on data sources, we can sometimes get different columns than what we have
            # cached. This is problematic but will require separate solution; don't fail for now
            # raise RuntimeError(f"Returned mismatch of columns: received: {df.columns}, expected: {self.columns}")
        return df

    def _to_dataframe(self) -> pandas.DataFrame:
        pandas_df = self.sql_executor.read_sql(self.sql_string)
        return pandas_df

    def to_snowpark(self) -> "snowflake.snowpark.DataFrame":
        snowpark_df = self.sql_executor.sql_to_snowpark(self.sql_string)
        return snowpark_df


# PandasExecNodes are responsible for taking in a pandas dataframe, performing some pandas operation and outputting a
# pandas dataframe


@attrs.frozen
class PandasExecNode:
    columns: Tuple[str]
    input_node: Optional[Union[PandasExecNode, SqlExecNode]]
    column_name_updater: Optional[Callable[[str], str]]  # Snowflake uses this method to uppercase all column names
    secret_resolver: Optional[SecretResolver]

    @classmethod
    def from_node_inputs(
        cls,
        query_node: node_interface.QueryNode,
        input_node: Optional[Union[PandasExecNode, SqlExecNode]],
        column_name_updater: Optional[Callable[[str], str]] = lambda x: x,
        secret_resolver: Optional[SecretResolver] = None,
    ) -> "PandasExecNode":
        kwargs = attrs.asdict(query_node, recurse=False)
        kwargs["input_node"] = input_node
        kwargs["columns"] = query_node.columns
        kwargs["column_name_updater"] = column_name_updater
        kwargs["secret_resolver"] = secret_resolver
        del kwargs["dialect"]
        del kwargs["compute_mode"]
        del kwargs["func"]
        del kwargs["node_id"]
        return cls(**kwargs)

    def to_dataframe(self) -> pandas.DataFrame:
        df = self._to_dataframe()
        if {c.lower() for c in df.columns} != {c.lower() for c in self.columns}:
            pass
            # Because we do not refresh schemas on data sources, we can sometimes get different columns than what we have
            # cached. This is problematic but will require separate solution; don't fail for now
            # raise RuntimeError(f"Returned mismatch of columns: received: {df.columns}, expected: {self.columns}")
        return df

    def to_arrow(self) -> pyarrow.RecordBatchReader:
        batch = pyarrow.RecordBatch.from_pandas(self.to_dataframe())
        return pyarrow.RecordBatchReader.from_batches(batch.schema, [batch])

    @abstractmethod
    def _to_dataframe(self) -> pandas.DataFrame:
        raise NotImplementedError



import base64
import datetime
import json
import os
import time
from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple

import boto3
import botocore.credentials
import botocore.errorfactory

from tecton_core import aws_credentials
from tecton_core.id_helper import IdHelper
from tecton_materialization.common.job_metadata import JobMetadataClient
from tecton_proto.common.aws_credentials__client_pb2 import AwsIamRole
from tecton_proto.common.id__client_pb2 import Id
from tecton_proto.materialization.job_metadata__client_pb2 import JobMetadata
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams


JOB_EXEC_PKEY_ATTRIBUTE = "id"
JOB_EXEC_LAST_UPDATED_ATTRIBUTE = "last_updated"
JOB_EXEC_DATA_ATTRIBUTE = "data"
JOB_EXEC_VERSION_ATTRIBUTE = "version"


class DynamoMetadataClient(JobMetadataClient):
    def __init__(self, table: str, attempt_id: Id, dynamodb_client: Any):
        self._table = table
        self._attempt_id = attempt_id
        self._dynamodb_client = dynamodb_client

    @staticmethod
    def for_params(materialization_task_params: MaterializationTaskParams) -> JobMetadataClient:
        return DynamoMetadataClient(
            table=materialization_task_params.job_metadata_table,
            attempt_id=materialization_task_params.attempt_id,
            dynamodb_client=_dynamodb_client(materialization_task_params),
        )

    @property
    def _key(self):
        return {JOB_EXEC_PKEY_ATTRIBUTE: {"S": IdHelper.to_string(self._attempt_id)}}

    def get(self) -> Tuple[JobMetadata, int]:
        item = self._dynamodb_client.get_item(
            TableName=self._table,
            Key=self._key,
            ConsistentRead=True,
        )["Item"]
        version = item[JOB_EXEC_VERSION_ATTRIBUTE]["N"]
        data = JobMetadata()
        data.ParseFromString(item[JOB_EXEC_DATA_ATTRIBUTE]["B"])
        return data, version

    def update(self, updater: Callable[[JobMetadata], Optional[JobMetadata]]) -> JobMetadata:
        num_retries = 100
        for i in range(num_retries):
            try:
                old_data, old_version = self.get()
                new_data = updater(old_data)
                if new_data is None:
                    return old_data
                now_seconds = int(time.time())
                self._dynamodb_client.put_item(
                    TableName=self._table,
                    Item={
                        **self._key,
                        JOB_EXEC_LAST_UPDATED_ATTRIBUTE: {"N": str(now_seconds)},
                        JOB_EXEC_DATA_ATTRIBUTE: {"B": new_data.SerializeToString()},
                        JOB_EXEC_VERSION_ATTRIBUTE: {"N": str(int(old_version) + 1)},
                    },
                    ConditionExpression="#version = :version",
                    ExpressionAttributeNames={"#version": JOB_EXEC_VERSION_ATTRIBUTE},
                    ExpressionAttributeValues={":version": {"N": str(old_version)}},
                )
                return new_data
            except botocore.errorfactory.ClientError as e:
                # Condition failed means we have a conflicting update
                if e.response["Error"]["Code"] == "ConditionalCheckFailedException" and i + 1 < num_retries:
                    continue
                else:
                    raise e


class DBFSCredentialFetcher(botocore.credentials.CachedCredentialFetcher):
    REFRESH_INTERVAL = datetime.timedelta(minutes=5)

    def __init__(self, path: str):
        # A fetch happens when expiration is expiry_window_seconds away, so this will cause us to re-fetch
        # credentials from DBFS every REFRESH_INTERVAL
        super().__init__(expiry_window_seconds=0)
        self._path = path

    def _create_cache_key(self) -> str:
        return "dbfs"

    def _get_credentials(self):
        with open(f"/dbfs{self._path}") as f:
            credentials = json.loads(base64.b64decode(f.read()))
            expiration = datetime.datetime.now(datetime.timezone.utc) + self.REFRESH_INTERVAL
            return {
                "Credentials": {
                    "AccessKeyId": credentials["accessKeyId"],
                    "SecretAccessKey": credentials["secretAccessKey"],
                    "SessionToken": credentials["sessionToken"],
                    "Expiration": expiration,
                },
            }


def _dynamo_session(params: MaterializationTaskParams) -> boto3.Session:
    if params.HasField("dynamodb_cross_account_role") or params.HasField("dynamodb_cross_account_role_arn"):
        assert not params.HasField("dbfs_credentials_path")
        role = AwsIamRole()
        if params.HasField("dynamodb_cross_account_role"):
            role = params.dynamodb_cross_account_role
        else:
            # TODO(meastham): Remove this fallback after dynamodb_cross_account_role support has been released in the
            #  backend for several releases
            role.role_arn = params.dynamodb_cross_account_role_arn
            if params.HasField("dynamodb_cross_account_external_id"):
                role.external_id = params.dynamodb_cross_account_external_id
        return aws_credentials.session_for_role(role, "tecton_materialization")
    elif params.HasField("dbfs_credentials_path"):
        return aws_credentials.session_for_fetcher(DBFSCredentialFetcher(params.dbfs_credentials_path), "dbfs")
    else:
        return boto3.Session()


def _dynamodb_client(params: MaterializationTaskParams):
    dynamo_params = {"region_name": params.dynamodb_table_region}
    if os.environ.get("TEST_ONLY_TECTON_DYNAMODB_ENDPOINT_OVERRIDE"):
        dynamo_params.update(endpoint_url=os.environ["TEST_ONLY_TECTON_DYNAMODB_ENDPOINT_OVERRIDE"])

    return _dynamo_session(params).client("dynamodb", **dynamo_params)



from __future__ import annotations

import inspect
import logging
import threading
import typing
from functools import wraps
from typing import Callable
from typing import Optional
from typing import TypeVar
from typing import Union
from typing import overload

import pendulum
import typeguard
from typeguard import typechecked
from typing_extensions import Concatenate
from typing_extensions import ParamSpec
from typing_extensions import TypeAlias

from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals.analytics import AnalyticsLogger
from tecton_core import conf
from tecton_core.errors import TectonAPIInaccessibleError
from tecton_core.errors import TectonAPIValidationError
from tecton_core.errors import TectonValidationError
from tecton_core.id_helper import IdHelper
from tecton_core.metadata_service_impl import trace as metadata_service_trace
from tecton_core.query.errors import UserDefinedTransformationError


_PT = ParamSpec("_PT")
_RT = TypeVar("_RT")

if typing.TYPE_CHECKING:
    import tecton.framework.base_tecton_object

    BaseTectonObject: TypeAlias = tecton.framework.base_tecton_object.BaseTectonObject


analytics = AnalyticsLogger()
thread_local_data = threading.local()
sdk_public_method_decorator_enabled = True


logger = logging.getLogger(__name__)


def disable_sdk_public_method_decorator():
    global sdk_public_method_decorator_enabled
    sdk_public_method_decorator_enabled = False


def enable_sdk_public_method_decorator():
    global sdk_public_method_decorator_enabled
    sdk_public_method_decorator_enabled = True


def documented_by(func: Callable) -> Callable[[Callable[_PT, _RT]], Callable[_PT, _RT]]:
    """Normally you should just use functools.wraps over this, but wraps() doesn't work nicely
    when you try to wrap a classmethod in a normal func.
    """

    @typeguard.typeguard_ignore
    def wrapper(target: Callable[_PT, _RT]) -> Callable[_PT, _RT]:
        target.__doc__ = func.__doc__
        return target

    return wrapper


@overload
def assert_remote_object(
    original_function: Callable[Concatenate[BaseTectonObject, _PT], _RT],
) -> Callable[Concatenate[BaseTectonObject, _PT], _RT]: ...


@overload
def assert_remote_object(
    *, error_message: Callable[[str], Exception] = errors.INVALID_USAGE_FOR_LOCAL_TECTON_OBJECT
) -> Callable[
    [Callable[Concatenate[BaseTectonObject, _PT], _RT]], Callable[Concatenate[BaseTectonObject, _PT], _RT]
]: ...


@typeguard.typeguard_ignore
def assert_remote_object(
    original_function: Optional[Callable[Concatenate[BaseTectonObject, _PT], _RT]] = None,
    *,
    error_message: Callable[[str], Exception] = errors.INVALID_USAGE_FOR_LOCAL_TECTON_OBJECT,
) -> Union[
    Callable[Concatenate[BaseTectonObject, _PT], _RT],
    Callable[[Callable[Concatenate[BaseTectonObject, _PT], _RT]], Callable[Concatenate[BaseTectonObject, _PT], _RT]],
]:
    """Assert this function is being called on a remote Tecton object, aka an object applied and fetched from the backend, and raise error otherwise.

    :param: error_message: error message to raise if the Tecton object is locally defined. The error_message param must contain a function that takes
    in the target function's name as a param and returns an Exception.
    """

    @typeguard.typeguard_ignore
    def inner_decorator(
        target: Callable[Concatenate[BaseTectonObject, _PT], _RT],
    ) -> Callable[Concatenate[BaseTectonObject, _PT], _RT]:
        @wraps(target)
        def wrapper(
            fco_object: BaseTectonObject,
            *args: _PT.args,
            **kwargs: _PT.kwargs,
        ) -> _RT:
            if fco_object.info._is_local_object:
                raise error_message(target.__name__)
            return target(fco_object, *args, **kwargs)

        return wrapper

    if original_function:
        return inner_decorator(original_function)

    return inner_decorator


@overload
def assert_local_object(
    original_function: Callable[Concatenate[BaseTectonObject, _PT], _RT],
) -> Callable[Concatenate[BaseTectonObject, _PT], _RT]: ...


@overload
def assert_local_object(
    *,
    error_message: Callable[[str], Exception] = errors.INVALID_USAGE_FOR_REMOTE_TECTON_OBJECT,
) -> Callable[
    [Callable[Concatenate[BaseTectonObject, _PT], _RT]], Callable[Concatenate[BaseTectonObject, _PT], _RT]
]: ...


@typeguard.typeguard_ignore
def assert_local_object(
    original_function: Optional[Callable[Concatenate[BaseTectonObject, _PT], _RT]] = None,
    *,
    error_message: Callable[[str], Exception] = errors.INVALID_USAGE_FOR_REMOTE_TECTON_OBJECT,
) -> Union[
    Callable[Concatenate[BaseTectonObject, _PT], _RT],
    Callable[[Callable[Concatenate[BaseTectonObject, _PT], _RT]], Callable[Concatenate[BaseTectonObject, _PT], _RT]],
]:
    """Assert this function is being called on a local Tecton object, aka an object created locally (as opposed to being fetched from the backend).

    :param: error_message: error message to raise if the Tecton object is not locally defined. The error_message param must contain a function that takes
    in the target function's name as a param and returns an Exception.
    """

    @typeguard.typeguard_ignore
    def inner_decorator(
        target: Callable[Concatenate[BaseTectonObject, _PT], _RT],
    ) -> Callable[Concatenate[BaseTectonObject, _PT], _RT]:
        @wraps(target)
        def wrapper(
            fco_object: BaseTectonObject,
            *args: _PT.args,
            **kwargs: _PT.kwargs,
        ) -> _RT:
            if not fco_object.info._is_local_object:
                raise error_message(target.__name__)
            return target(fco_object, *args, **kwargs)

        return wrapper

    if original_function:
        return inner_decorator(original_function)

    return inner_decorator


@overload
def sdk_public_method(original_function: Callable[_PT, _RT]) -> Callable[_PT, _RT]: ...


@overload
def sdk_public_method() -> Callable[[Callable[_PT, _RT]], Callable[_PT, _RT]]: ...


@typeguard.typeguard_ignore
def sdk_public_method(
    original_function: Optional[Callable[_PT, _RT]] = None,
) -> Union[Callable[_PT, _RT], Callable[[Callable[_PT, _RT]], Callable[_PT, _RT]]]:
    """Decorator for public SDK methods that should have analytics logging.

    :param original_function: The function to be wrapped.
    """

    @typeguard.typeguard_ignore
    def inner_decorator(target_func: Callable[_PT, _RT]) -> Callable[_PT, _RT]:
        arg_names = _get_arg_names(target_func)

        @wraps(target_func)
        def _sdk_public_method_wrapper(*args: _PT.args, **kwargs: _PT.kwargs) -> _RT:
            if not sdk_public_method_decorator_enabled:
                return target_func(*args, **kwargs)

            if not hasattr(thread_local_data, "in_tecton_sdk_public_method_wrapper"):
                thread_local_data.in_tecton_sdk_public_method_wrapper = False

            already_in_wrapper = thread_local_data.in_tecton_sdk_public_method_wrapper
            if already_in_wrapper:
                return target_func(*args, **kwargs)

            try:
                thread_local_data.in_tecton_sdk_public_method_wrapper = True
                return _invoke_and_transform_errors(
                    target_func, args, kwargs, arg_names, analytics is not None, not already_in_wrapper, typecheck=True
                )
            finally:
                thread_local_data.in_tecton_sdk_public_method_wrapper = already_in_wrapper

        return _sdk_public_method_wrapper

    if original_function:
        return inner_decorator(original_function)

    return inner_decorator


def _is_method(func):
    params = inspect.signature(func).parameters
    return "self" in params or "cls" in params


def _get_arg_names(func):
    arg_names = []
    for i, param in enumerate(inspect.signature(func).parameters.values()):
        name = param.name
        if i == 0 and _is_method(func):
            continue
        arg_names.append(name)
    return arg_names


def _invoke_and_transform_errors(
    func, args, kwargs, arg_names, log_analytics: bool, is_top_level: bool, typecheck: bool = False
):
    if typecheck:
        func = typechecked(func)

    if _is_method(func):
        method_object = args[0]
        method_args = args[1:]
    else:
        method_object = None
        method_args = args

    trace_id = IdHelper.generate_string_id()

    if log_analytics:
        start_time = pendulum.now("UTC")
        analytics.log_method_entry(trace_id, method_object, func, method_args, kwargs, arg_names)

    metadata_service_trace.set_trace_id(trace_id)
    return_value = None

    # It's important that this is initialized, because some exceptions (e.g. KeyboardInterrupt) do not inherit from
    # Exception because they are not meant to be caught.
    caught_exception = None
    try:
        return_value = func(*args, **kwargs)
        caught_exception = None
        return return_value
    except TectonAPIValidationError as e:
        caught_exception = e
        raise errors.VALIDATION_ERROR_FROM_MDS(str(e), trace_id)
    except TectonAPIInaccessibleError as e:
        caught_exception = e
        raise errors.MDS_INACCESSIBLE(metadata_service._get_host_port())
    except (UserDefinedTransformationError, TectonValidationError) as e:
        caught_exception = e
        # for user-caused errors we want to minimize the traceback for cleaner output.
        # nested exception (ie, exception from UDF) though still might have some meaningful traceback.
        if not conf.get_bool("TECTON_DEBUG") and e.can_drop_traceback:
            raise e.with_traceback(None)
        raise
    except Exception as e:
        caught_exception = e
        # Do not chain the exception, just let it pass through. This leads to more legible stack traces and errors.
        raise
    finally:
        metadata_service_trace.set_trace_id(None)
        if log_analytics:
            end_time = pendulum.now("UTC")
            execution_time = end_time - start_time
            analytics.log_method_return(trace_id, method_object, func, return_value, execution_time, caught_exception)



from typing import Any
from typing import Dict
from typing import List
from typing import Union

import pandas
from pyspark.sql import DataFrame

from tecton._internals import errors
from tecton.framework import data_frame
from tecton_core import pipeline_common
from tecton_core import specs
from tecton_core.compute_mode import ComputeMode
from tecton_core.errors import TectonValidationError
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.query.dialect import Dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.nodes import UserSpecifiedDataNode
from tecton_proto.common.data_source_type__client_pb2 import DataSourceType


def convert_mock_inputs_to_mock_sources(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fdw: FeatureDefinitionWrapper,
    mock_inputs: Dict[str, Union[pandas.DataFrame, DataFrame]],
) -> Dict[str, NodeRef]:
    validate_mock_inputs(mock_inputs, fdw)

    input_ds_id_map = pipeline_common.get_input_name_to_ds_id_map(fdw.pipeline)

    mock_data_sources = {
        input_ds_id_map[key]: NodeRef(
            UserSpecifiedDataNode(dialect, compute_mode, data_frame.TectonDataFrame._create(mock_inputs[key]))
        )
        for key in mock_inputs.keys()
    }
    return mock_data_sources


def validate_mock_inputs(mock_inputs: Dict[str, Union[pandas.DataFrame, DataFrame]], fd: FeatureDefinitionWrapper):
    """Validate the mock data source data used for `run()` and `test_run()`.

    This validation does not enforce that the mock data schema is exactly equivalent to the data source schema, however
    it does check for schemas that will usually lead to a bad downstream error.

    For example, it is okay if the mock data is missing some fields (the fields may not be used/needed in the feature
    view transformation), but it's not okay if the mock data is missing the `timestamp_field` defined in the data
    source, since that is used by Tecton time filtering.
    """
    # Validate that mock_inputs keys are a subset of data sources.
    expected_input_names = [
        node.data_source_node.input_name for node in pipeline_common.get_all_data_source_nodes(fd.pipeline)
    ]
    mock_inputs_keys = set(mock_inputs.keys())
    if not mock_inputs_keys.issubset(expected_input_names):
        raise errors.FV_INVALID_MOCK_INPUTS(list(mock_inputs_keys), expected_input_names)

    # Validate all push no batch sources are a subset of mock_inputs keys
    expected_input_name_to_push_no_batch_ds = {
        input_name: ds
        for input_name, ds in (
            (input_name, fd.get_data_source_with_input_name(input_name)) for input_name in expected_input_names
        )
        if ds.type == DataSourceType.PUSH_NO_BATCH
    }
    expected_push_no_batch_input_names = set(expected_input_name_to_push_no_batch_ds.keys())
    missing_push_no_batch_sources = expected_push_no_batch_input_names.difference(mock_inputs.keys())
    if len(missing_push_no_batch_sources) > 0:
        msg = f"Stream sources without batch sources must be provided as mock inputs. Please add mock inputs: {missing_push_no_batch_sources}."
        raise TectonValidationError(msg)

    # Validate some required fields in the mock data schemas.
    for key, mock_df in mock_inputs.items():
        data_source = fd.get_data_source_with_input_name(key)
        column_names = get_pandas_or_spark_df_or_dict_columns(mock_df)

        # Note: We are intentionally not checking for missing or extraneous columns in the data sources. For more
        # complex data sources with many columns it is reasonable that users may only want to provide a subset of
        # the columns as inputs

        if data_source.type != DataSourceType.PUSH_NO_BATCH:
            # PUSH_NO_BATCH data sources have no batch source to validate
            validate_mock_batch_data_source(data_source, column_names)


def validate_mock_batch_data_source(data_source: specs.DataSourceSpec, column_names: List[str]):
    if (
        data_source.batch_source.timestamp_field is not None
        and data_source.batch_source.timestamp_field not in column_names
    ):
        msg = f"Data Source '{data_source.name}' specified a timestamp_field '{data_source.batch_source.timestamp_field}', but that timestamp field was not found in the mock data source columns: '{column_names}'"
        raise TectonValidationError(msg)

    if isinstance(data_source.batch_source, specs.HiveSourceSpec):
        for datetime_partition_column in data_source.batch_source.datetime_partition_columns:
            if datetime_partition_column.column_name not in column_names:
                msg = f"Data Source '{data_source.name}' specified a datetime partition column '{datetime_partition_column.column_name}', but that column was not found in the mock data source schema: '{column_names}'"
                raise errors.TectonValidationError(msg)


def get_pandas_or_spark_df_or_dict_columns(df: Union[pandas.DataFrame, DataFrame, Dict[str, Any]]) -> List[str]:
    if isinstance(df, pandas.DataFrame):
        return list(df.columns)
    elif isinstance(df, DataFrame):
        return [field.name for field in df.schema.fields]
    elif isinstance(df, dict):
        return list(df.keys())
    else:
        msg = f"Unexpected mock data type '{type(df)}'. Should be a Pandas or Spark data frame or Python dictionary."
        raise TypeError(msg)



import logging
from datetime import datetime
from typing import Dict
from typing import Optional

import attrs
import pendulum
import pyspark

from tecton_core import specs
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.query.node_interface import DataframeWrapper
from tecton_core.query.node_interface import NodeRef
from tecton_proto.args.pipeline__client_pb2 import DataSourceNode
from tecton_spark import data_source_helper
from tecton_spark import offline_store
from tecton_spark.query.node import SparkExecNode


logger = logging.getLogger(__name__)


@attrs.frozen
class UserSpecifiedDataSparkNode(SparkExecNode):
    data: DataframeWrapper
    metadata: Optional[Dict[str, any]]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        return self.data.to_spark()


@attrs.frozen
class DataSparkNode(SparkExecNode):
    data: DataframeWrapper
    metadata: Optional[Dict[str, any]]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        return self.data.to_spark()


@attrs.frozen
class MockDataSourceScanSparkNode(SparkExecNode):
    data: SparkExecNode
    ds: specs.DataSourceSpec
    start_time: Optional[datetime]
    end_time: Optional[datetime]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        df = self.data.to_dataframe(spark)
        if self.start_time or self.end_time:
            df = data_source_helper.apply_partition_and_timestamp_filter(
                df, self.ds.batch_source, self.start_time, self.end_time
            )
        return df


@attrs.frozen
class DataSourceScanSparkNode(SparkExecNode):
    ds: specs.DataSourceSpec
    is_stream: bool
    start_time: Optional[datetime]
    end_time: Optional[datetime]
    ds_node: Optional[DataSourceNode]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        df = data_source_helper.get_ds_dataframe(
            spark,
            self.ds,
            consume_streaming_data_source=self.is_stream,
            start_time=self.start_time,
            end_time=self.end_time,
        )
        return df


# This is used for debugging method.
@attrs.frozen
class RawDataSourceScanSparkNode(SparkExecNode):
    ds: specs.DataSourceSpec

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        df = data_source_helper.get_non_dsf_raw_dataframe(spark, self.ds.batch_source)
        return df


@attrs.frozen
class OfflineStoreScanSparkNode(SparkExecNode):
    feature_definition_wrapper: FeatureDefinitionWrapper
    partition_time_filter: Optional[pendulum.Period]
    # TODO: pushdown join keys filter based on the provided spine (currently this is not used by Spark implementation)
    entity_filter: Optional[NodeRef] = None

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        offline_reader = offline_store.get_offline_store_reader(spark, self.feature_definition_wrapper)
        try:
            # None implies no timestamp filtering. When we implement time filter pushdown, it will go here
            df = offline_reader.read(self.partition_time_filter)
            return df
        except pyspark.sql.utils.AnalysisException as e:
            logger.warning(
                f"Failed to read from the Offline Store. Please ensure that materialization backfills have completed for Feature View '{self.feature_definition_wrapper.name}' or set `from_source=True`."
            )
            raise e



"""Helper module for creating spark data source operations such as creating DataFrames and
registering temp views. Methods within this class should exclusively operate on proto
representations of the data model.
"""

import json
import os
import time
from datetime import datetime
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

import pendulum
from pyspark.sql import functions
from pyspark.sql.functions import coalesce
from pyspark.sql.functions import to_timestamp
from pyspark.sql.types import ArrayType
from pyspark.sql.types import DataType
from pyspark.sql.types import DateType
from pyspark.sql.types import MapType
from pyspark.sql.types import StringType
from pyspark.sql.types import StructField
from pyspark.sql.types import StructType
from typeguard import typechecked

import tecton_spark.errors_spark
from tecton_core import conf
from tecton_core import specs
from tecton_core.filter_context import FilterContext
from tecton_proto.args.data_source_config__client_pb2 import INITIAL_STREAM_POSITION_LATEST
from tecton_proto.args.data_source_config__client_pb2 import INITIAL_STREAM_POSITION_TRIM_HORIZON
from tecton_proto.args.data_source_config__client_pb2 import INITIAL_STREAM_POSITION_UNSPECIFIED
from tecton_proto.common.data_source_type__client_pb2 import DataSourceType
from tecton_proto.data.batch_data_source__client_pb2 import FileDataSourceFormat
from tecton_spark.data_source_credentials import get_kafka_secrets
from tecton_spark.spark_schema_wrapper import SparkSchemaWrapper
from tecton_spark.type_annotations import PySparkColumn
from tecton_spark.type_annotations import PySparkDataFrame
from tecton_spark.type_annotations import PySparkSession


INITIAL_STREAM_POSITION_STR_TO_ENUM = {
    "latest": INITIAL_STREAM_POSITION_LATEST,
    "trim_horizon": INITIAL_STREAM_POSITION_TRIM_HORIZON,
}

INITIAL_STREAM_POSITION_ENUM_TO_STR: Dict[str, Optional[str]] = {
    v: k for k, v in INITIAL_STREAM_POSITION_STR_TO_ENUM.items()
}
INITIAL_STREAM_POSITION_ENUM_TO_STR[INITIAL_STREAM_POSITION_UNSPECIFIED] = None

KAFKA_DEFAULT_MAX_OFFSETS_PER_TRIGGER = 100000

KAFKA_STARTING_OFFSET_CONFIG_KEYS = {"startingOffsetsByTimestamp", "startingOffsets"}

# Set of temporal Kafka consumer settings that will be set by the environment variables set in the
# cluster's `textproto` files.
#
# Defines maximum number of records (across all partitions) from Kafka to be consumed by a micro-batch.
KAFKA_MAX_OFFSETS_PER_TRIGGER_ENV = "KAFKA_MAX_OFFSETS_PER_TRIGGER"
# Option to set starting timestamps for each Kafka partitions so that all of the Kafka's retention
# data is not processed. It's only relevant for the fist streaming job before the checkpoint exists on S3.
KAFKA_STARTING_OFFSETS_NUM_HOURS_IN_THE_PAST_ENV = "KAFKA_STARTING_OFFSETS_NUM_HOURS_IN_THE_PAST"
# Only necessary when KAFKA_STARTING_OFFSETS_NUM_HOURS_IN_THE_PAST_ENV is set.
KAFKA_NUM_PARTITIONS_ENV = "KAFKA_NUM_PARTITIONS"

TEST_ONLY_UNITY_CATALOG_NAME = "TECTON_LOCAL_INTEGRATION_TEST_ONLY_UNITY_CATALOG"


def _is_running_on_emr() -> bool:
    return (
        "EMR_RELEASE_LABEL" in os.environ
        or os.environ.get("TECTON_RUNTIME_ENV") == "EMR"
        or conf.get_or_none("TECTON_RUNTIME_ENV") == "EMR"
    )


def _get_raw_hive_table_dataframe(spark: PySparkSession, database: str, table: str) -> PySparkDataFrame:
    spark.sql("USE {}".format(database))
    return spark.table(table)


def _get_raw_unity_table_dataframe(spark: PySparkSession, catalog: str, schema: str, table: str) -> PySparkDataFrame:
    # USE CATALOG is only supported in databricks sql but not local pyspark
    # so we'd have to tweak for local integration tests to use "USE" sql statement
    # this means customers won't be able to use a unity catalog named "TECTON_LOCAL_INTEGRATION_TEST_ONLY_UNITY_CATALOG"
    if catalog == TEST_ONLY_UNITY_CATALOG_NAME:
        spark.sql(f"USE {catalog}")
    else:
        spark.sql(f"USE CATALOG {catalog}")
    spark.sql(f"USE {schema}")
    return spark.table(table)


@typechecked
def get_non_dsf_raw_dataframe(
    spark: PySparkSession,
    data_source: specs.BatchSourceSpec,
    called_for_schema_computation: bool = False,
) -> PySparkDataFrame:
    """Returns a DataFrame of the raw, untranslated data defined by the given BatchDataSource proto.

    :param spark: Spark session.
    :param data_source: BatchDataSource proto. BatchDataSource must not be a data source function (spark_batch_config).
    :param called_for_schema_computation: If set, optimizations are applied for faster schema computations.
                                          i.e. FileDSConfig.schema_uri is used to avoid expensive partition discovery.

    :return: The DataFrame for the raw data.
    """

    assert not isinstance(
        data_source, specs.SparkBatchSourceSpec
    ), "get_raw_dataframe can not be used with data source function (spark_batch_config)."

    if isinstance(data_source, specs.HiveSourceSpec):
        df = _get_raw_hive_table_dataframe(spark, data_source.database, data_source.table)
    elif isinstance(data_source, specs.UnitySourceSpec):
        df = _get_raw_unity_table_dataframe(spark, data_source.catalog, data_source.schema, data_source.table)
    elif isinstance(data_source, specs.RedshiftSourceSpec):
        df = get_redshift_dataframe(
            spark,
            data_source.endpoint,
            data_source.temp_s3,
            data_source.table,
            data_source.query,
        )
    elif isinstance(data_source, specs.SnowflakeSourceSpec):
        df = get_snowflake_dataframe(
            spark,
            data_source.url,
            data_source.database,
            data_source.schema,
            data_source.warehouse,
            data_source.role,
            data_source.table,
            data_source.query,
        )
    elif isinstance(data_source, specs.FileSourceSpec):
        # FileDataSource
        reader = spark.read
        uri = data_source.uri
        if called_for_schema_computation and data_source.schema_uri:
            # Setting basePath includes the path-based partitions in the DataFrame schema.
            # https://spark.apache.org/docs/latest/sql-data-sources-parquet.html#partition-discovery
            reader = reader.option("basePath", data_source.uri)
            uri = data_source.schema_uri

        if data_source.schema_override:
            schema = SparkSchemaWrapper.from_proto(data_source.schema_override)
            reader = reader.schema(schema.unwrap())

        if data_source.file_format == FileDataSourceFormat.FILE_DATA_SOURCE_FORMAT_JSON:

            def action():
                return reader.json(uri)

        elif data_source.file_format == FileDataSourceFormat.FILE_DATA_SOURCE_FORMAT_PARQUET:

            def action():
                return reader.parquet(uri)

        elif data_source.file_format == FileDataSourceFormat.FILE_DATA_SOURCE_FORMAT_CSV:

            def action():
                return reader.csv(uri, header=True)

        else:
            msg = f"Unsupported file format '{data_source.file_format}'"
            raise AssertionError(msg)

        df = tecton_spark.errors_spark.handleDataAccessErrors(action, data_source.uri)
        if data_source.convert_to_glue_format:
            df = convert_json_like_schema_to_glue_format(spark, df)
    elif isinstance(data_source, specs.PushTableSourceSpec):
        df = spark.read.format("delta").load(data_source.ingested_data_location)
    else:
        msg = f"Unexpected data source type for source {data_source}"
        raise ValueError(msg)

    return df


@typechecked
def get_table_dataframe(
    spark: PySparkSession,
    data_source: specs.BatchSourceSpec,
    called_for_schema_computation: bool = False,
) -> PySparkDataFrame:
    """Returns a DataFrame for a table defined by given BatchDataSource proto.

    :param spark: Spark session.
    :param data_source: BatchDataSource proto. BatchDataSource must not be a data source function (spark_batch_config).
    :param called_for_schema_computation: If set, optimizations are applied for faster schema computations.
                                          i.e. FileDSConfig.schema_uri is used to avoid expensive partition discovery.

    :return: The DataFrame created from the data source.
    """
    assert not isinstance(
        data_source, specs.SparkBatchSourceSpec
    ), f"get_table_dataframe can not be used with data source function (spark_batch_config). Data source: {data_source}"

    df = get_non_dsf_raw_dataframe(spark, data_source, called_for_schema_computation)
    if data_source.post_processor:
        df = data_source.post_processor(df)
    if data_source.timestamp_field:
        df = apply_timestamp_column(df, data_source.timestamp_field, data_source.timestamp_format)

    return df


def get_redshift_dataframe(
    spark: PySparkSession, endpoint: str, temp_s3: str, table: Optional[str] = None, query: Optional[str] = None
) -> PySparkDataFrame:
    """Returns a DataFrame for a Redshift table defined by given RedshiftDataSource proto.

    :param table: The table name in redshift
    :param temp_s3: The s3 URI for temp data
    :param endpoint: The connection endpoint for redshift (without user or password)
    :param spark: Spark session.

    :return: The DataFrame created from the data source.
    """

    if _is_running_on_emr():
        spark_format = "io.github.spark_redshift_community.spark.redshift"
    else:
        spark_format = "com.databricks.spark.redshift"

    params = {"user": conf.get_or_raise("REDSHIFT_USER"), "password": conf.get_or_raise("REDSHIFT_PASSWORD")}
    full_connection_string = f"jdbc:redshift://{endpoint};user={params['user']};password={params['password']}"

    df_reader = (
        spark.read.format(spark_format)
        .option("url", full_connection_string)
        .option("tempdir", temp_s3)
        .option("forward_spark_s3_credentials", "true")
    )

    if table and query:
        msg = "Should only specify one of table and query sources for redshift"
        raise AssertionError(msg)
    if not table and not query:
        msg = "Missing both table and query sources for redshift, exactly one must be present"
        raise AssertionError(msg)

    if table:
        df_reader = df_reader.option("dbtable", table)
    else:
        df_reader = df_reader.option("query", query)

    df = df_reader.load()
    return df


def get_snowflake_dataframe(
    spark: PySparkSession,
    url: str,
    database: str,
    schema: str,
    warehouse: str,
    role: Optional[str] = None,
    table: Optional[str] = None,
    query: Optional[str] = None,
    spark_schema: Optional[StructType] = None,
) -> PySparkDataFrame:
    """Returns a Spark DataFrame from a Snowflake table or query.

    Casts the DataFrame to match spark_schema if it is specified.

    Requires the Spark-Snowflake connector JAR to be installed. See https://docs.snowflake.com/en/user-guide/spark-connector.
    """

    if (table and query) or (not table and not query):
        msg = "Exactly one of table and query must be specified."
        raise ValueError(msg)

    user = conf.get_or_none("SNOWFLAKE_USER")
    password = conf.get_or_none("SNOWFLAKE_PASSWORD")
    if not user or not password:
        msg = "Snowflake user and password not configured. Instructions at https://docs.tecton.ai/docs/setting-up-tecton/connecting-data-sources/connecting-to-snowflake-using-spark"
        raise ValueError(msg)

    options = {
        "sfUrl": url,
        "sfUser": user,
        "sfPassword": password,
        "sfDatabase": database,
        "sfSchema": schema,
        "sfWarehouse": warehouse,
        "APPLICATION": "tecton-ai",
    }

    if role:
        options["sfRole"] = role

    df_reader = spark.read.format("snowflake").options(**options)

    if table:
        df_reader = df_reader.option("dbtable", table)
    else:
        df_reader = df_reader.option("query", query)

    df = df_reader.load()
    if spark_schema is not None:
        for field in spark_schema:
            df = df.withColumn(field.name, functions.col(field.name).cast(field.dataType))
    return df


def apply_timestamp_column(df: PySparkDataFrame, ts_column: str, ts_format: Optional[str]) -> PySparkDataFrame:
    # Verify the raw source's timestamp column is of type "string"
    column_names = df.schema.names
    if ts_column not in column_names:
        msg = f"Timestamp Column '{ts_column}' not found in schema. Found: {column_names}"
        raise AssertionError(msg)

    ts_type = df.schema[ts_column].dataType.jsonValue()
    if ts_type != "timestamp":
        assert (
            ts_type == "string"
        ), f"Timestamp Column '{ts_column}' has type '{ts_type}', expected 'string' or 'timestamp'"
        # Apply timestamp transform

        # Here we use coalesce to first try transforming string to timestamp using the user provided format,
        # and if it doesn't work we'll instead let Spark figure it out.
        # Ideally, if the user provided format didn't work, we would not fallback to the Spark default. However, it
        # would be difficult to remove this behavior, and it's hard to imagine a scenario where this would be a problem
        # other than being a bit too "magical". Ref: https://tecton.atlassian.net/browse/TEC-6611
        df = df.withColumn(ts_column, coalesce(to_timestamp(df[ts_column], ts_format), to_timestamp(df[ts_column])))

    return df


@typechecked
def apply_partition_and_timestamp_filter(
    df: PySparkDataFrame,
    batch_source: specs.BatchSourceSpec,
    start_time: Optional[Union[pendulum.DateTime, datetime]] = None,
    end_time: Optional[Union[pendulum.DateTime, datetime]] = None,
) -> PySparkDataFrame:
    """Applies a partition and timestamp filters if the respective column names are set.

    :return: The DataFrame with filter applied.
    """

    # add datetime partition filters
    if isinstance(batch_source, specs.HiveSourceSpec) and batch_source.datetime_partition_columns:
        partition_filter = _build_partition_filter(batch_source.datetime_partition_columns, start_time, end_time)
        if partition_filter is not None:
            df = df.where(partition_filter)

    # add timestamp filter
    if batch_source.timestamp_field:
        ts_column = functions.col(batch_source.timestamp_field)
        if start_time:
            df = df.where(ts_column >= start_time)
        if end_time:
            df = df.where(ts_column < end_time)

    return df


# Generate filter on time_range with at most OR of 2 filters.
# This means we could end up scanning more partitions than necessary, but number extra scanned will be at most
# 3x.
# Worst case: time_range = 367 days across 3 years, we scan entire 3 years.
#             time_range = 365 days, we scan up to 2 years including that 1 year range
#             time_range = 28.1 days, we could scan all of January + February + March = 90 days
#
# 2 cases to consider
# Example: partition cols y, m, d, h: in both examples, the time range is between day and month, so we don't add any
# filters on hour, even though it would be possible to scan some fewer partitions if we did
# (A)Time range: 2020 Jan 10 10:00:00 AM - 2020 Jan 15 5:00:00 AM
#  --> (y = start.year & m = start.month & d >= start.day & d <= end.day)
# (B)Time range: 2019 Dec 21 10:00:00 AM - 2020 Jan 10 5:00:00 AM
#  --> ((y = start.year & m = start.month & d >= start.day) | (y = end.year & m = end.month & d <= end.day))
def _build_partition_filter(
    datetime_partition_columns: specs.DatetimePartitionColumnSpec,
    start_time: Optional[Union[pendulum.DateTime, datetime]] = None,
    end_time: Optional[Union[pendulum.DateTime, datetime]] = None,
) -> PySparkColumn:
    if not start_time and not end_time:
        return None

    range_duration = end_time - start_time if start_time and end_time else None

    # Filters relating to start_time and end_time
    start_filters = []
    end_filters = []
    # Common filter that applies to the entire range
    common_filters = []

    # sort partition columns by the number of seconds they represent from highest to lowest
    partitions_high_to_low = sorted(datetime_partition_columns, key=lambda c: c.minimum_seconds, reverse=True)
    for partition in partitions_high_to_low:
        partition_col = functions.col(partition.column_name)
        partition_value_at_start = _partition_value_for_time(partition, start_time) if start_time else None
        partition_value_at_end = _partition_value_for_time(partition, end_time) if end_time else None
        # If partition's datepart minimum length is >= range_duration secs, we can be sure that 2 or fewer equality filters are enough to cover all possible times in time_limits
        # If range_duration is None/unbounded, always use range filter
        if range_duration and partition.minimum_seconds >= range_duration.total_seconds():
            if partition_value_at_start == partition_value_at_end:
                common_filters.append(partition_col == partition_value_at_start)
            else:
                start_filters.append(partition_col == partition_value_at_start)
                end_filters.append(partition_col == partition_value_at_end)
        # Otherwise, we need to use a range filter
        else:
            start_range_filter = partition_col >= partition_value_at_start if partition_value_at_start else None
            end_range_filter = partition_col <= partition_value_at_end if partition_value_at_end else None

            # Case A: there are only common filters
            if len(start_filters) == 0:
                if start_range_filter is not None:
                    common_filters.append(start_range_filter)
                if end_range_filter is not None:
                    common_filters.append(end_range_filter)
                # we can't combine range filters on multiple columns, so break and ignore any smaller columns
                break
            # Case B
            else:
                if start_range_filter is not None:
                    start_filters.append(start_range_filter)
                if end_range_filter is not None:
                    end_filters.append(end_range_filter)
                # we can't combine range filters on multiple columns, so break and ignore any smaller columns
                break

    common_filter = _and_filters_in_list(common_filters)
    start_filter = _and_filters_in_list(start_filters)
    end_filter = _and_filters_in_list(end_filters)
    return common_filter & (start_filter | end_filter)


def _partition_value_for_time(partition, time):
    fmt = partition.format_string
    # On mac/linux strftime supports these formats, but
    # Windows python does not
    # Zero-padded formats are safe for a string comparison. Otherwise we need to compare ints
    # As long as the values returned here are ints, the column will be implicitly converted if needed.
    if fmt == "%-Y":
        return time.year
    elif fmt == "%-m":
        return time.month
    elif fmt == "%-d":
        return time.day
    elif fmt == "%-H":
        return time.hour
    return time.strftime(fmt)


def _and_filters_in_list(filter_list):
    if len(filter_list) == 0:
        return functions.lit(True)
    else:
        from functools import reduce

        return reduce(lambda x, y: x & y, filter_list)


def create_kinesis_stream_reader(
    spark: PySparkSession,
    stream_source: specs.KinesisSourceSpec,
    option_overrides: Optional[Dict[str, str]],
) -> PySparkDataFrame:
    """
    Returns a DataFrame representing a Kinesis stream reader.

    :param option_overrides: Spark options that should override options set implicitly (e.g. ``stream_name``) or
        explicitly  (e.g. ``options``) by the data source definition.
    """
    options = {"streamName": stream_source.stream_name}
    initial_stream_position = _get_initial_stream_position(stream_source)
    if _is_running_on_emr():
        options.update(
            {
                "endpointUrl": f"https://kinesis.{stream_source.region}.amazonaws.com",
                "kinesis.client.describeShardInterval": "30s",
                "startingPosition": initial_stream_position,
            }
        )
    else:
        options.update(
            {"region": stream_source.region, "shardFetchInterval": "30s", "initialPosition": initial_stream_position}
        )

    databricks_to_qubole_map = {
        "awsaccesskey": "awsAccessKeyId",
        "rolearn": "awsSTSRoleARN",
        "rolesessionname": "awsSTSSessionName",
    }
    lowercase_data_source_options = {option.key.lower(): option.value for option in stream_source.options}
    for option in stream_source.options:
        if option.key.lower() in databricks_to_qubole_map and _is_running_on_emr():
            if option.key.lower() == "rolearn" and "rolesessionname" not in lowercase_data_source_options:
                # this field must be supplied if we use roleArn for qubole kinesis reader
                options["awsSTSSessionName"] = "tecton-materialization"
            options[databricks_to_qubole_map[option.key.lower()]] = option.value
        else:
            options[option.key] = option.value

    if option_overrides:
        options.update(option_overrides)

    reader = spark.readStream.format("kinesis").options(**options)
    return reader.load()


def create_kafka_stream_reader(
    spark: PySparkSession,
    stream_source: specs.KafkaSourceSpec,
    option_overrides: Optional[Dict[str, str]],
) -> PySparkDataFrame:
    """Returns a Kafka stream reader.

    :param data_source_options: Spark options specified in the data source definition.
    :param option_overrides: Spark options that should override options set implicitly (e.g. ``topics``) or explicitly
        (e.g. ``options``) by the data source definition.
    """
    options = {o.key: o.value for o in stream_source.options}
    options["kafka.bootstrap.servers"] = stream_source.bootstrap_servers
    options["subscribe"] = stream_source.topics
    # Kafka by default consumes all the exisitng data into a single micro-batch, that can overwhelm
    # the Spark cluster during backfilling from a stream for the first time.
    # Set the default number of records to be read per micro-batch (across all partitions).
    options["maxOffsetsPerTrigger"] = str(KAFKA_DEFAULT_MAX_OFFSETS_PER_TRIGGER)
    if all(key not in options for key in KAFKA_STARTING_OFFSET_CONFIG_KEYS):
        # Don't override startingOffsets if it or similar option is set
        # explicitly in the data source definition.
        options["startingOffsets"] = "earliest"
    options = _populate_kafka_consumer_options(stream_source.topics, options)

    if stream_source.ssl_keystore_location:
        local_keystore_loc, local_keystore_password = get_kafka_secrets(
            stream_source.ssl_keystore_location, stream_source.ssl_keystore_password_secret_id
        )
        options["kafka.ssl.keystore.location"] = local_keystore_loc
        if local_keystore_password:
            options["kafka.ssl.keystore.password"] = local_keystore_password
    if stream_source.ssl_truststore_location:
        local_truststore_loc, local_truststore_password = get_kafka_secrets(
            stream_source.ssl_truststore_location, stream_source.ssl_truststore_password_secret_id
        )
        options["kafka.ssl.truststore.location"] = local_truststore_loc
        if local_truststore_password:
            options["kafka.ssl.truststore.password"] = local_truststore_password
    if stream_source.security_protocol:
        options["kafka.security.protocol"] = stream_source.security_protocol

        # Hack: dynamic PLAIN SASL_SSL authentication when the
        # authentication is unset. Approved usage for Square until Data Source
        # Functions is launched.
        # TODO(TEC-9976): remove once Square is on Data Source Functions.
        if (
            stream_source.security_protocol == "SASL_SSL"
            and options.get("kafka.sasl.mechanism") == "PLAIN"
            and options.get("kafka.sasl.jaas.config") is None
        ):
            sasl_username = conf.get_or_none("SECRET_TECTON_KAFKA_SASL_USERNAME")
            sasl_password = conf.get_or_none("SECRET_TECTON_KAFKA_SASL_PASSWORD")
            kafka_sasl_jaas_config = f"kafkashaded.org.apache.kafka.common.security.plain.PlainLoginModule required username='{sasl_username}' password='{sasl_password}';"
            options["kafka.sasl.jaas.config"] = kafka_sasl_jaas_config

    if option_overrides:
        options.update(option_overrides)

    reader = spark.readStream.format("kafka").options(**options)
    return reader.load()


# This is a short term fix to unblock Kafka customers. We will replace these with proper FCO configurations.
# This function sets 2 Kafka consumer options `maxOffsetsPerTrigger` and `startingOffsetsByTimestamp` that
# are gatekept by the Environment variables set into the Spark cluster.
#  - The `maxOffsetsPerTrigger` represents the max number of records read into a SSS micro-batch.
#  - The `startingOffsetsByTimestamp` dynamically overrides the starting timestamps on the Kafka partitions
#    to limit the amount of data read from the Kafka data source.
def _populate_kafka_consumer_options(topics: str, kafka_options: Dict[str, Any]) -> Dict[str, Any]:
    max_offsets = os.environ.get(KAFKA_MAX_OFFSETS_PER_TRIGGER_ENV)
    if max_offsets is not None:
        assert str.isdigit(max_offsets), f"{KAFKA_MAX_OFFSETS_PER_TRIGGER_ENV} must be a string encoded integer"
        kafka_options["maxOffsetsPerTrigger"] = max_offsets

    num_hours_to_process = os.environ.get(KAFKA_STARTING_OFFSETS_NUM_HOURS_IN_THE_PAST_ENV)
    if num_hours_to_process is not None:
        assert str.isdigit(
            num_hours_to_process
        ), f"{KAFKA_STARTING_OFFSETS_NUM_HOURS_IN_THE_PAST_ENV} must be a string encoded integer"
        num_partitions = os.environ.get(KAFKA_NUM_PARTITIONS_ENV)
        assert num_partitions is not None and str.isdigit(
            num_partitions
        ), f"{KAFKA_NUM_PARTITIONS_ENV} missing/invalid when {KAFKA_STARTING_OFFSETS_NUM_HOURS_IN_THE_PAST_ENV} is set"
        # `startingOffsetsByTimestamp` needs unix timestamps in millis per partition number.
        current_ts = int(time.time())
        starting_ts = (current_ts - int(num_hours_to_process) * 3600) * 1000
        ts_per_partition = {str(partition): starting_ts for partition in range(int(num_partitions))}
        topic_names = topics.split(",")
        starting_offsets_by_ts = {topic_name: ts_per_partition for topic_name in topic_names}
        kafka_options["startingOffsetsByTimestamp"] = json.dumps(starting_offsets_by_ts)

    return kafka_options


def get_non_dsf_raw_stream_dataframe(
    spark: PySparkSession, stream_data_source: specs.StreamSourceSpec, option_overrides: Optional[Dict[str, str]] = None
) -> PySparkDataFrame:
    """Returns a DataFrame representing the raw stream data.

    :param spark: Spark session.
    :param stream_data_source: StreamDataSource proto. StreamDataSource must not be a data source function (spark_stream_config).
    :param option_overrides: A dictionary of Spark readStream options that will override any readStream options set by
        the data source.

    :return: The DataFrame for the raw stream data source.
    """
    if isinstance(stream_data_source, specs.KinesisSourceSpec):
        df = create_kinesis_stream_reader(
            spark,
            stream_data_source,
            option_overrides,
        )
    elif isinstance(stream_data_source, specs.KafkaSourceSpec):
        df = create_kafka_stream_reader(
            spark,
            stream_data_source,
            option_overrides,
        )
    else:
        msg = f"Unknown stream data source type: {stream_data_source}"
        raise ValueError(msg)
    return df


def get_stream_dataframe(
    spark: PySparkSession, stream_data_source: specs.StreamSourceSpec, option_overrides: Optional[Dict[str, str]] = None
) -> PySparkDataFrame:
    """Returns a DataFrame representing a stream data source *without* any options specified.
    Use get_stream_dataframe_with_options to get a DataFrame with stream-specific options.

    :param spark: Spark session.
    :param stream_data_source: StreamDataSource proto. StreamDataSource must not be a data source function (spark_stream_config).
    :param option_overrides: A dictionary of Spark readStream options that will override any readStream options set by
        the data source.

    :return: The DataFrame created from the data source.
    """
    assert isinstance(
        stream_data_source, (specs.KinesisSourceSpec, specs.KafkaSourceSpec)
    ), "get_stream_dataframe can not be used with data source function (spark_stream_config)."

    df = get_non_dsf_raw_stream_dataframe(spark, stream_data_source, option_overrides)
    return stream_data_source.post_processor(df)


def get_stream_dataframe_with_options(
    spark: PySparkSession, stream_data_source: specs.StreamSourceSpec, option_overrides: Optional[Dict[str, str]] = None
) -> PySparkDataFrame:
    """Returns a DataFrame representing a stream data source with additional options:
        - drop duplicate column names
        - initial stream position

    :param spark: Spark session.
    :param stream_data_source: StreamDataSource proto. StreamDataSource must not be a data source function (spark_stream_config).
    :param option_overrides: A dictionary of Spark readStream options that will override any readStream options set by
        the data source.

    :return: The DataFrame created from the data source.
    """
    assert not isinstance(
        stream_data_source, specs.SparkStreamSourceSpec
    ), "get_stream_dataframe_with_options can not be used with data source function (spark_stream_config)."

    df = get_stream_dataframe(spark, stream_data_source, option_overrides)

    dedup_columns = list(stream_data_source.deduplication_column_names)
    if dedup_columns:
        df = df.dropDuplicates(dedup_columns)

    return df


def _get_initial_stream_position(stream_data_source: specs.StreamSourceSpec) -> Optional[str]:
    """Returns initial stream position as a string (e.g. "latest") for the streaming data source.

    :param stream_data_source: StreamDataSource proto.

    :return: The initial stream position in string format.
    """
    if isinstance(stream_data_source, specs.KinesisSourceSpec):
        return INITIAL_STREAM_POSITION_ENUM_TO_STR[stream_data_source.initial_stream_position]
    else:
        return None


@typechecked
def get_ds_dataframe(
    spark: PySparkSession,
    data_source: specs.DataSourceSpec,
    consume_streaming_data_source: bool,
    start_time: Optional[Union[pendulum.DateTime, datetime]] = None,
    end_time: Optional[Union[pendulum.DateTime, datetime]] = None,
    called_for_schema_computation: bool = False,
    stream_option_overrides: Optional[Dict[str, str]] = None,
) -> PySparkDataFrame:
    if consume_streaming_data_source and (start_time or end_time):
        msg = "Can't specify start or end time when consuming streaming data source"
        raise AssertionError(msg)

    if consume_streaming_data_source:
        assert (
            data_source.stream_source
        ), f"Can't consume streaming data source from the data source: {data_source.name}."

        if isinstance(data_source.stream_source, specs.SparkStreamSourceSpec):
            df = data_source.stream_source.function(spark)
        else:
            df = get_stream_dataframe_with_options(spark, data_source.stream_source, stream_option_overrides)
    else:
        if isinstance(data_source.batch_source, specs.SparkBatchSourceSpec):
            df = get_data_source_function_batch_dataframe(
                spark, data_source.batch_source, start_time=start_time, end_time=end_time
            )
        else:
            df = get_table_dataframe(
                spark, data_source.batch_source, called_for_schema_computation=called_for_schema_computation
            )
            if start_time or end_time:
                df = apply_partition_and_timestamp_filter(df, data_source.batch_source, start_time, end_time)

        if data_source.type == DataSourceType.STREAM_WITH_BATCH:
            schema = data_source.stream_source.spark_schema
            # Since 0.8, data source schema is optionally derived. If all feature views that depends on a data source have `run_transformation_validation=False`, the data source schema is not derived.
            if schema.fields:
                cols = [field.name for field in schema.fields]
                df = df.select(*cols)
        elif data_source.type in (DataSourceType.PUSH_NO_BATCH, DataSourceType.PUSH_WITH_BATCH):
            schema = data_source.schema.tecton_schema
            cols = [f"`{c.name}`" for c in schema.columns]
            df = df.select(*cols)

    return df


def convert_json_like_schema_to_glue_format(spark: PySparkSession, df: PySparkDataFrame) -> PySparkDataFrame:
    """
    Converts a DataFrame schema to lowercase. This assumes JSON so
    MapTypes or Arrays of non-StructTypes are not allowed.

    :param spark: Spark session.
    :param df: DataFrame input.
    :return: DataFrame with lowercase schema.
    """

    def _get_lowercase_schema(datatype: DataType) -> DataType:
        if type(datatype) == ArrayType:
            return _get_lowercase_array_schema(datatype)
        elif type(datatype) == StructType:
            return _get_lowercase_structtype_schema(datatype)
        elif type(col.dataType) == MapType:
            msg = "MapType not supported in JSON schema"
            raise TypeError(msg)
        return datatype

    def _get_lowercase_structtype_schema(s: StructType) -> StructType:
        assert type(s) == StructType, f"Invalid argument type {type(s)}, expected StructType"
        struct_fields = []
        for col in s:
            datatype = _get_lowercase_schema(col.dataType)
            struct_fields.append(StructField(col.name.lower(), datatype))
        return StructType(struct_fields)

    def _get_lowercase_array_schema(c: ArrayType) -> ArrayType:
        assert (
            type(c.elementType) == StructType
        ), f"Invalid ArrayType element type {type(c)}, expected StructType for valid JSON arrays."
        datatype = c.elementType
        struct_schema = _get_lowercase_structtype_schema(datatype)
        return ArrayType(struct_schema)

    # Simple columns (LongType, StringType, etc) can just be renamed without
    # casting schema.
    # Nested fields within complex columns (ArrayType, StructType) must also be recursively converted
    # to lowercase names, so they must be casted.
    # DateType columns should be converted to StringType to match Glue schemas.
    new_fields = []
    for col in df.schema:
        if type(col.dataType) in [ArrayType, StructType, MapType]:
            t = _get_lowercase_schema(col.dataType)
            new_fields.append(functions.col(col.name).cast(t).alias(col.name.lower()))
        elif type(col.dataType) is DateType:
            new_fields.append(functions.col(col.name).cast(StringType()).alias(col.name.lower()))
        else:
            new_fields.append(functions.col(col.name).alias(col.name.lower()))
    return df.select(new_fields)


def get_data_source_function_batch_dataframe(
    spark: PySparkSession,
    data_source: specs.SparkBatchSourceSpec,
    start_time: Optional[datetime],
    end_time: Optional[datetime],
) -> PySparkDataFrame:
    if data_source.supports_time_filtering:
        filter_context = FilterContext(start_time, end_time)
        return data_source.function(spark, filter_context)
    return data_source.function(spark)



import contextlib
from typing import List
from typing import Tuple

import pyarrow

from tecton_core import offline_store
from tecton_core.compute_mode import ComputeMode
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.query.builder import build_materialization_querytree
from tecton_core.query.dialect import Dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.query_tree_executor import QueryTreeExecutor
from tecton_materialization.common.task_params import TimeInterval
from tecton_materialization.common.task_params import feature_definition_from_task_params
from tecton_materialization.common.task_params import job_query_intervals
from tecton_materialization.ray.job_status import JobStatusClient
from tecton_materialization.ray.materialization_utils import get_delta_writer
from tecton_materialization.ray.materialization_utils import write_to_online_store
from tecton_materialization.ray.nodes import AddTimePartitionNode
from tecton_proto.materialization.job_metadata__client_pb2 import TectonManagedStage
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams
from tecton_proto.offlinestore.delta import metadata__client_pb2 as metadata_pb2


PYARROW_ASC = "ascending"
PYARROW_DESC = "descending"


def run_batch_materialization(
    materialization_task_params: MaterializationTaskParams,
    job_status_client: JobStatusClient,
    executor: QueryTreeExecutor,
):
    intervals = job_query_intervals(materialization_task_params)
    for idx, interval in enumerate(intervals):
        job_status_client.set_query_index(idx, len(intervals))
        materialize_interval(
            interval=interval,
            materialization_task_params=materialization_task_params,
            job_status_client=job_status_client,
            executor=executor,
        )


def materialize_interval(
    interval: TimeInterval,
    materialization_task_params: MaterializationTaskParams,
    job_status_client: JobStatusClient,
    executor: QueryTreeExecutor,
):
    fd = feature_definition_from_task_params(materialization_task_params)
    assert fd.writes_to_offline_store, f"Offline materialization is required for FeatureView {fd.id} ({fd.name})"
    assert fd.has_delta_offline_store, f"Delta is required for FeatureView {fd.id} ({fd.name})"

    qt = _get_batch_materialization_plan(fd, interval)
    materialized_data = executor.exec_qt(qt).result_table

    # Sorting rows withing batches helps improve writing parquet files: fewer partitions are written in parallel.
    # Also, secondary sorting by join keys can improve reading performance (if filter by join key will be pushed down to arrow reader).
    materialized_data = sort_rows_in_batches(
        materialized_data,
        by=[(offline_store.TIME_PARTITION, PYARROW_ASC), *[(key, PYARROW_ASC) for key in fd.join_keys]],
    )

    should_write_to_online_store = (
        materialization_task_params.batch_task_info.batch_parameters.write_to_online_feature_store
    )

    offline_stage_monitor = job_status_client.create_stage_monitor(
        TectonManagedStage.StageType.OFFLINE_STORE,
        "Unload features to offline store",
    )
    online_stage_monitor = (
        job_status_client.create_stage_monitor(
            TectonManagedStage.StageType.ONLINE_STORE,
            "Unload features to online store",
        )
        if should_write_to_online_store
        else None
    )

    is_overwrite = materialization_task_params.batch_task_info.batch_parameters.is_overwrite

    delta_writer = get_delta_writer(materialization_task_params)
    parts = None
    with offline_stage_monitor():
        transaction_metadata = metadata_pb2.TectonDeltaMetadata()
        transaction_metadata.feature_start_time.FromDatetime(interval.start)
        transaction_exists = delta_writer.transaction_exists(transaction_metadata)
        if not is_overwrite and transaction_exists:
            print(
                f"Found previous commit with metadata {transaction_metadata} for data in range {interval.start} - {interval.end}. Skipping writing to delta table."
            )
        else:

            @delta_writer.transaction(transaction_metadata)
            def txn() -> List[str]:
                if is_overwrite:
                    delta_writer.delete_time_range(interval)
                return delta_writer.write(materialized_data)

            parts = txn()

    if should_write_to_online_store:
        with online_stage_monitor(), contextlib.ExitStack() as stack:
            if parts is None:
                # We skipped the txn because of matching metadata, but we still need to write out the parquet files for
                # the online store writer. We can accomplish this by using write() and then later abort() to delete
                # the files when we're done with them.
                parts = delta_writer.write(materialized_data)
                stack.callback(delta_writer.abort)

            # TODO(meastham): Probably should send these all at once to the online store copier
            for uri in parts:
                write_to_online_store(
                    materialization_task_params.online_store_writer_config,
                    materialization_task_params.feature_view,
                    materialization_task_params.batch_task_info.batch_parameters.feature_end_time,
                    fd,
                    uri,
                )


def _get_batch_materialization_plan(fd: FeatureDefinitionWrapper, interval: TimeInterval) -> NodeRef:
    tree = build_materialization_querytree(
        dialect=Dialect.DUCKDB,
        compute_mode=ComputeMode.RIFT,
        fdw=fd,
        for_stream=False,
        feature_data_time_limits=interval.to_pendulum(),
    )
    return AddTimePartitionNode.for_feature_definition(fd, tree)


def sort_rows_in_batches(reader: pyarrow.RecordBatchReader, by: List[Tuple[str, str]]) -> pyarrow.RecordBatchReader:
    """
    Create new RecordBatchReader with rows sorted within each batch.

    :param reader: iterator over record batches
    :param by: sorting conditions - list of tuple(name, order)
    """

    def batch_iter():
        while True:
            try:
                next_batch = reader.read_next_batch()
            except StopIteration:
                return

            yield next_batch.sort_by(by)

    return pyarrow.RecordBatchReader.from_batches(reader.schema, batch_iter())



import base64
import enum
import hashlib
import http.server as http_server
import secrets
import socketserver
import sys
import time
from typing import Optional
from typing import Tuple
from urllib.parse import parse_qs
from urllib.parse import urlencode
from urllib.parse import urlparse

import attrs
import click
import requests

from tecton_core import conf
from tecton_core import http


AUTH_SERVER = "https://login.tecton.ai/oauth2/default/.well-known/oauth-authorization-server"
USER_INFO_ENDPOINT = "https://login.tecton.ai/oauth2/default/v1/userinfo"

# this must match the uri that the okta application was configured with
OKTA_EXPECTED_PORTS = [10003, 10013, 10023]

# our http server will populate this when it receives the callback from okta
_CALLBACK_PATH = None


@attrs.frozen
class UserProfile:
    name: str
    email: str
    id: str


class OktaCallbackReceivingServer(http_server.BaseHTTPRequestHandler):
    def log_message(self, format, *args):
        pass

    def do_GET(self):
        global _CALLBACK_PATH
        _CALLBACK_PATH = self.path
        self.send_response(301)
        self.send_header("Location", "https://www.tecton.ai/authentication-success")
        self.end_headers()


def get_metadata():
    try:
        response = http.session().get(AUTH_SERVER)
        response.raise_for_status()
    except requests.RequestException as e:
        raise SystemExit(e)
    return response.json()


# Okta docs: https://developer.okta.com/docs/guides/refresh-tokens/use-refresh-token/
class RefreshFlow:
    def __init__(self):
        self.metadata = get_metadata()

    def get_tokens(self, refresh_token):
        params = {
            "refresh_token": refresh_token,
            "grant_type": "refresh_token",
            "client_id": conf.get_or_none("CLI_CLIENT_ID"),
            "scope": "openid profile email",
        }
        try:
            return get_tokens_helper(self.metadata["token_endpoint"], params)
        except requests.RequestException as e:
            msg = "Authorization token expired. Please reauthenticate with `tecton login`"
            raise PermissionError(msg)


class AuthFlowType(enum.Enum):
    # Spin up local http server that can intercept the okta callback and avoid having to
    # copy & paste
    BROWSER_HANDS_FREE = 1

    # User manually opens link to login, and has to copy & paste the code.
    # Also used for notebook login (tecton.login())
    # Relies in Python's built in input() function
    BROWSER_MANUAL = 2

    # Tecton-internal session token authorization.
    SESSION_TOKEN = 3

    # Pyspark notebooks cannot use input() so BROWSER_MANUAL will not work.
    # Used for 2 step notebook login (tecton.login(interactive=False) + tecton.login_with_code())
    NOTEBOOK_2_STEP = 4


# Okta docs: https://developer.okta.com/docs/guides/implement-auth-code-pkce/use-flow/
class OktaAuthorizationFlow:
    def __init__(
        self, auth_flow_type: AuthFlowType = AuthFlowType.BROWSER_HANDS_FREE, okta_session_token: Optional[str] = None
    ):
        self.metadata = get_metadata()
        self.auth_flow_type = auth_flow_type

        self.okta_session_token = okta_session_token

    def build_authorization_url(
        self,
        code_verifier: str,
        redirect_uri: str,
        cli_client_id: str,
    ) -> Tuple[str, str, str]:
        # base64's `urlsafe_b64encode` uses '=' as padding. These are not URL safe when used in URL paramaters.
        code_challenge = base64.urlsafe_b64encode(hashlib.sha256(code_verifier.encode("utf-8")).digest()).rstrip(b"=")

        state = str(secrets.randbits(64))

        params = {
            "response_type": "code",
            "client_id": cli_client_id,
            "redirect_uri": redirect_uri,
            "state": state,
            "scope": "openid offline_access profile email",
            "code_challenge_method": "S256",
            "code_challenge": code_challenge,
        }
        if self.auth_flow_type == AuthFlowType.SESSION_TOKEN:
            params["sessionToken"] = self.okta_session_token
            params["response_mode"] = "query"
        authorize_url = f"{self.metadata['authorization_endpoint']}?{urlencode(params)}"
        return authorize_url, state

    def _browser_auth(self, code_verifier: str, cli_client_id: str) -> Tuple[str, str]:
        e: Optional[OSError] = None
        for port in OKTA_EXPECTED_PORTS:
            try:
                httpd = socketserver.TCPServer(("", int(port)), OktaCallbackReceivingServer)
                break
            except OSError as e:
                # socket in use, try other port
                if e.errno == 48:
                    continue
                else:
                    print(
                        "Encountered error with authorization callback. Your environment may not support automatic login; try running `tecton login --manual` instead.",
                        file=sys.stderr,
                    )
                    sys.exit(1)
        else:
            if e:
                raise e

        redirect_uri = f"http://localhost:{port}/authorization/callback"
        authorize_url, state = self.build_authorization_url(code_verifier, redirect_uri, cli_client_id)
        print("Requesting authorization for Tecton CLI via browser. ")
        # Sleep before opening the browser to improve the UX and make it less jarring.
        time.sleep(1)
        try:
            print("If browser doesn't open automatically, use `tecton login --manual`")
            click.launch(authorize_url)
            httpd.handle_request()
        finally:
            httpd.server_close()

        return self._extract_code_from_url(_CALLBACK_PATH, state), redirect_uri

    def _extract_code_from_url(self, url: str, state: str) -> str:
        parsed = parse_qs(urlparse(url).query)
        if "error" in parsed:
            print(f"Encountered error: {parsed['error']}", file=sys.stderr)
            if "access_denied" in parsed["error"]:
                print(
                    "Please contact your Tecton administrator to verify you have been granted cluster access.",
                    file=sys.stderr,
                )
            sys.exit(1)
        code = parsed["code"][0]
        assert state == parsed["state"][0]
        from tecton.cli import printer

        printer.safe_print("✅ Authentication successful!")
        return code

    def _manual_auth(self, authorize_url):
        print("Please visit the following link to login and access the authentication code:")
        print(f"{authorize_url}")
        code = input("Paste authentication code here and press [Enter]:").strip()
        return code

    # NOTE: this is for internal use only. Requires an Okta session token.
    # This works by leveraging the Okta session token to skip browser-based
    # Okta authentication. The parameters are stored in the redirect uri from
    # Okta. https://developer.okta.com/docs/reference/api/oidc/#authorize
    def _session_token_auth(self, authorize_url: str, state: str) -> str:
        # We disallow redirects so that we can extract the code from the
        # redirect location.
        r = http.session().get(authorize_url, allow_redirects=False)
        r.raise_for_status()
        if r.status_code != 302:
            print("Unsuccessful headless Okta auth.", file=sys.stderr)
            sys.exit(1)
        url = r.headers.get("location")
        return self._extract_code_from_url(url, state)

    # To receive the auth code, we send the user to their browser to login to okta.
    # Once they've logged in, Okta will redirect them to a locally served webpage with the auth code.
    # Then, this function will return that auth code.
    def get_authorization_code(self, cli_client_id: str) -> Tuple[str, str, str]:
        code_verifier = secrets.token_urlsafe(43)

        if self.auth_flow_type == AuthFlowType.BROWSER_HANDS_FREE:
            code, redirect_uri = self._browser_auth(code_verifier, cli_client_id)
            return code, code_verifier, redirect_uri

        code_verifier, authorize_url, redirect_uri, state = self.generate_pkce_and_login_link(cli_client_id)

        if self.auth_flow_type == AuthFlowType.SESSION_TOKEN:
            code = self._session_token_auth(authorize_url, state)
        elif self.auth_flow_type == AuthFlowType.BROWSER_MANUAL:
            code = self._manual_auth(authorize_url)
        else:
            # NOTEBOOK_2_STEP gets the authorization code via manual user input
            msg = "AuthFlowType.NOTEBOOK_2_STEP should use generate_pkce_and_login_link"
            raise NotImplementedError(msg)
        return code, code_verifier, redirect_uri

    def generate_pkce_and_login_link(self, cli_client_id) -> Tuple[str, str, str, str]:
        """
        Generates the PKCE for OAuth: https://developer.okta.com/docs/guides/implement-grant-type/authcodepkce/main/#create-the-proof-key-for-code-exchange
        And /authorize URL and params needed to request an auth code: https://developer.okta.com/docs/guides/implement-grant-type/authcodepkce/main/#create-the-proof-key-for-code-exchange

        :return: Tuple of:
            1. code_verifier
            2. authorize_url
            3. redirect_uri
            4. state
            which are documented in the above 2 links
        """
        if self.auth_flow_type == AuthFlowType.BROWSER_HANDS_FREE:
            msg = "Not supported for AuthFlowType.BROWSER_HANDS_FREE."
            raise NotImplementedError(msg)

        code_verifier = secrets.token_urlsafe(43)
        redirect_uri = "https://www.tecton.ai/authorization-callback"
        authorize_url, state = self.build_authorization_url(code_verifier, redirect_uri, cli_client_id)

        return code_verifier, authorize_url, redirect_uri, state

    def get_tokens(self, auth_code, code_verifier, redirect_uri, cli_client_id):
        params = {
            "grant_type": "authorization_code",
            "code": auth_code,
            "code_verifier": code_verifier,
            "redirect_uri": redirect_uri,
            "client_id": cli_client_id,
        }
        try:
            return get_tokens_helper(self.metadata["token_endpoint"], params)
        except Exception as e:
            raise SystemExit(e)


def get_tokens_helper(endpoint, params):
    response = http.session().post(endpoint, data=params)
    response.raise_for_status()

    response_json = response.json()
    access_token = response_json.get("access_token", None)
    id_token = response_json.get("id_token", None)
    refresh_token = response_json.get("refresh_token", None)
    expiration = time.time() + int(response_json.get("expires_in"))
    return access_token, id_token, refresh_token, expiration


# returns None if neither access nor refresh token are found in config
def get_token_refresh_if_needed():
    token = conf.get_or_none("OAUTH_ACCESS_TOKEN")
    if not token:
        return None
    expiration = conf.get_or_none("OAUTH_ACCESS_TOKEN_EXPIRATION")
    if expiration and time.time() < int(expiration):
        return token
    else:
        f = RefreshFlow()
        refresh_token = conf.get_or_none("OAUTH_REFRESH_TOKEN")
        if not refresh_token:
            return None
        access_token, _, _, expiration = f.get_tokens(refresh_token)
        conf.set_okta_tokens(access_token, expiration, refresh_token)
        return access_token


def get_user_profile() -> Optional[UserProfile]:
    token = get_token_refresh_if_needed()

    if not token:
        return None

    try:
        header = {"Authorization": f"Bearer {token}"}
        response = http.session().get(USER_INFO_ENDPOINT, headers=header)
        response.raise_for_status()
    except requests.RequestException as e:
        print("Unable to get user info. Please try logging in again using `tecton login`")
        raise SystemExit(e)

    response = response.json()
    return UserProfile(
        name=response.get("name", ""),
        email=response.get("email", ""),
        id=response.get("sub", ""),
    )


if __name__ == "__main__":
    # what does this do?
    f = OktaAuthorizationFlow()
    code, verifier, redirect_uri = f.get_authorization_code()
    access_token, id_token, refresh_token, expiration = f.get_tokens(code, verifier, redirect_uri)

    rf = RefreshFlow()
    print(rf.get_tokens(refresh_token))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/auth/authorization_service__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.auditlog import metadata__client_pb2 as tecton__proto_dot_auditlog_dot_metadata____client__pb2
from tecton_proto.auth import principal__client_pb2 as tecton__proto_dot_auth_dot_principal____client__pb2
from tecton_proto.auth import resource__client_pb2 as tecton__proto_dot_auth_dot_resource____client__pb2
from tecton_proto.auth import resource_role_assignments__client_pb2 as tecton__proto_dot_auth_dot_resource__role__assignments____client__pb2
from tecton_proto.auth import service__client_pb2 as tecton__proto_dot_auth_dot_service____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n5tecton_proto/auth/authorization_service__client.proto\x12\x11tecton_proto.auth\x1a,tecton_proto/auditlog/metadata__client.proto\x1a)tecton_proto/auth/principal__client.proto\x1a(tecton_proto/auth/resource__client.proto\x1a\x39tecton_proto/auth/resource_role_assignments__client.proto\x1a\'tecton_proto/auth/service__client.proto\"\x11\n\x0fGetRolesRequest\"K\n\x10GetRolesResponse\x12\x37\n\x05roles\x18\x01 \x03(\x0b\x32!.tecton_proto.auth.RoleDefinitionR\x05roles\"\x85\x03\n\x0eRoleDefinition\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12 \n\x0b\x64\x65scription\x18\x03 \x01(\tR\x0b\x64\x65scription\x12`\n\x1c\x61ssignable_on_resource_types\x18\x04 \x03(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x19\x61ssignableOnResourceTypes\x12\x63\n\x1d\x61ssignable_to_principal_types\x18\x05 \x03(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\x1a\x61ssignableToPrincipalTypes\x12I\n\x0bpermissions\x18\x06 \x03(\x0b\x32\'.tecton_proto.auth.PermissionDefinitionR\x0bpermissions\x12\x1b\n\tlegacy_id\x18\x07 \x01(\tR\x08legacyId\"m\n\x14PermissionDefinition\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12 \n\x0b\x64\x65scription\x18\x02 \x01(\tR\x0b\x64\x65scription\x12#\n\ris_authorized\x18\x03 \x01(\x08R\x0cisAuthorized\"\x84\x02\n\x17GetAssignedRolesRequest\x12G\n\x0eprincipal_type\x18\x01 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\rprincipalType\x12!\n\x0cprincipal_id\x18\x02 \x01(\tR\x0bprincipalId\x12\x44\n\rresource_type\x18\x03 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12!\n\x0cresource_ids\x18\x04 \x03(\tR\x0bresourceIds\x12\x14\n\x05roles\x18\x05 \x03(\tR\x05roles\"k\n\x18GetAssignedRolesResponse\x12O\n\x0b\x61ssignments\x18\x01 \x03(\x0b\x32-.tecton_proto.auth.ResourceAndRoleAssignmentsR\x0b\x61ssignments\"\x85\x02\n\x18ListAssignedRolesRequest\x12G\n\x0eprincipal_type\x18\x01 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\rprincipalType\x12!\n\x0cprincipal_id\x18\x02 \x01(\tR\x0bprincipalId\x12\x44\n\rresource_type\x18\x03 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12!\n\x0cresource_ids\x18\x04 \x03(\tR\x0bresourceIds\x12\x14\n\x05roles\x18\x05 \x03(\tR\x05roles\"n\n\x19ListAssignedRolesResponse\x12Q\n\x0b\x61ssignments\x18\x01 \x03(\x0b\x32/.tecton_proto.auth.ResourceAndRoleAssignmentsV2R\x0b\x61ssignments\"\xc5\x01\n\x16GetIsAuthorizedRequest\x12G\n\x0eprincipal_type\x18\x01 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\rprincipalType\x12!\n\x0cprincipal_id\x18\x02 \x01(\tR\x0bprincipalId\x12?\n\x0bpermissions\x18\x03 \x03(\x0b\x32\x1d.tecton_proto.auth.PermissionR\x0bpermissions\"Z\n\x17GetIsAuthorizedResponse\x12?\n\x0bpermissions\x18\x01 \x03(\x0b\x32\x1d.tecton_proto.auth.PermissionR\x0bpermissions\"^\n\x12\x41ssignRolesRequest\x12H\n\x0b\x61ssignments\x18\x01 \x03(\x0b\x32\x1d.tecton_proto.auth.AssignmentB\x07\xea\x99\xbd\x46\x02\x08\x02R\x0b\x61ssignments\"\x15\n\x13\x41ssignRolesResponse\"`\n\x14UnassignRolesRequest\x12H\n\x0b\x61ssignments\x18\x01 \x03(\x0b\x32\x1d.tecton_proto.auth.AssignmentB\x07\xea\x99\xbd\x46\x02\x08\x02R\x0b\x61ssignments\"\x17\n\x15UnassignRolesResponse\"\xa7\x02\n\nAssignment\x12M\n\rresource_type\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeB\x07\xea\x99\xbd\x46\x02\x08\x02R\x0cresourceType\x12(\n\x0bresource_id\x18\x02 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\nresourceId\x12\"\n\x04role\x18\x03 \x01(\tB\x0e\xea\x99\xbd\x46\x02\x08\x02\xea\x99\xbd\x46\x02\x10\x01R\x04role\x12P\n\x0eprincipal_type\x18\x04 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeB\x07\xea\x99\xbd\x46\x02\x08\x02R\rprincipalType\x12*\n\x0cprincipal_id\x18\x05 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x0bprincipalId\"\xb4\x02\n\x15\x41ssignRolesPutRequest\x12M\n\rresource_type\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeB\x07\xea\x99\xbd\x46\x02\x08\x02R\x0cresourceType\x12(\n\x0bresource_id\x18\x02 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\nresourceId\x12P\n\x0eprincipal_type\x18\x03 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeB\x07\xea\x99\xbd\x46\x02\x08\x02R\rprincipalType\x12*\n\x0cprincipal_id\x18\x04 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x0bprincipalId\x12$\n\x05roles\x18\x05 \x03(\tB\x0e\xea\x99\xbd\x46\x02\x08\x02\xea\x99\xbd\x46\x02\x10\x01R\x05roles\"\x18\n\x16\x41ssignRolesPutResponse\"\xe9\x01\n\x1dGetAuthorizedResourcesRequest\x12G\n\x0eprincipal_type\x18\x01 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\rprincipalType\x12!\n\x0cprincipal_id\x18\x02 \x01(\tR\x0bprincipalId\x12\x44\n\rresource_type\x18\x03 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12\x16\n\x06\x61\x63tion\x18\x04 \x01(\tR\x06\x61\x63tion\"{\n\x1eGetAuthorizedResourcesResponse\x12Y\n\x14\x61uthorized_resources\x18\x01 \x03(\x0b\x32&.tecton_proto.auth.AuthorizedResourcesR\x13\x61uthorizedResources\"P\n\x13\x41uthorizedResources\x12\x1f\n\x0bresource_id\x18\x01 \x01(\tR\nresourceId\x12\x18\n\x07\x61\x63tions\x18\x02 \x03(\tR\x07\x61\x63tions\"\xe6\x01\n\x1cGetAssignedPrincipalsRequest\x12\x44\n\rresource_type\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12\x1f\n\x0bresource_id\x18\x02 \x01(\tR\nresourceId\x12\x14\n\x05roles\x18\x03 \x03(\tR\x05roles\x12I\n\x0fprincipal_types\x18\x04 \x03(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\x0eprincipalTypes\"e\n\x1dGetAssignedPrincipalsResponse\x12\x44\n\x0b\x61ssignments\x18\x01 \x03(\x0b\x32\".tecton_proto.auth.AssignmentBasicR\x0b\x61ssignments\"\xe7\x01\n\x1dListAssignedPrincipalsRequest\x12\x44\n\rresource_type\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12\x1f\n\x0bresource_id\x18\x02 \x01(\tR\nresourceId\x12\x14\n\x05roles\x18\x03 \x03(\tR\x05roles\x12I\n\x0fprincipal_types\x18\x04 \x03(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\x0eprincipalTypes\"h\n\x1eListAssignedPrincipalsResponse\x12\x46\n\x0b\x61ssignments\x18\x01 \x03(\x0b\x32$.tecton_proto.auth.AssignmentBasicV2R\x0b\x61ssignments\"\xc2\x01\n\x0f\x41ssignmentBasic\x12?\n\tprincipal\x18\x01 \x01(\x0b\x32!.tecton_proto.auth.PrincipalBasicR\tprincipal\x12\x14\n\x05roles\x18\x02 \x03(\tR\x05roles\x12X\n\x10role_assignments\x18\x03 \x03(\x0b\x32-.tecton_proto.auth.ResourceAndRoleAssignmentsR\x0froleAssignments\"\xb0\x01\n\x11\x41ssignmentBasicV2\x12?\n\tprincipal\x18\x01 \x01(\x0b\x32!.tecton_proto.auth.PrincipalBasicR\tprincipal\x12Z\n\x10role_assignments\x18\x02 \x03(\x0b\x32/.tecton_proto.auth.ResourceAndRoleAssignmentsV2R\x0froleAssignments\"\xee\x01\n\x18GetAppPermissionsRequest\x12G\n\x0eprincipal_type\x18\x01 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\rprincipalType\x12!\n\x0cprincipal_id\x18\x02 \x01(\tR\x0bprincipalId\x12\x66\n\x19resource_type_permissions\x18\x05 \x03(\x0b\x32*.tecton_proto.auth.ResourceTypePermissionsR\x17resourceTypePermissions\"y\n\x17ResourceTypePermissions\x12\x44\n\rresource_type\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12\x18\n\x07\x61\x63tions\x18\x02 \x03(\tR\x07\x61\x63tions\"\x93\x01\n\x19GetAppPermissionsResponse\x12v\n\x1fresource_type_permission_values\x18\x01 \x03(\x0b\x32/.tecton_proto.auth.ResourceTypePermissionValuesR\x1cresourceTypePermissionValues\"\xb5\x01\n\x1cResourceTypePermissionValues\x12\x44\n\rresource_type\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12O\n\x11permission_values\x18\x02 \x03(\x0b\x32\".tecton_proto.auth.PermissionValueR\x10permissionValues\"\x8b\x01\n\nPermission\x12\x44\n\rresource_type\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12\x1f\n\x0bresource_id\x18\x02 \x01(\tR\nresourceId\x12\x16\n\x06\x61\x63tion\x18\x03 \x01(\tR\x06\x61\x63tion\"N\n\x0fPermissionValue\x12\x16\n\x06\x61\x63tion\x18\x01 \x01(\tR\x06\x61\x63tion\x12#\n\ris_authorized\x18\x02 \x01(\x08R\x0cisAuthorized\"\x97\x02\n\x1eGetWorkspacePermissionsRequest\x12G\n\x0eprincipal_type\x18\x01 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\rprincipalType\x12!\n\x0cprincipal_id\x18\x02 \x01(\tR\x0bprincipalId\x12!\n\x0cworkspace_id\x18\x03 \x01(\tR\x0bworkspaceId\x12\x66\n\x19resource_type_permissions\x18\x04 \x03(\x0b\x32*.tecton_proto.auth.ResourceTypePermissionsR\x17resourceTypePermissions\"\x99\x01\n\x1fGetWorkspacePermissionsResponse\x12v\n\x1fresource_type_permission_values\x18\x01 \x03(\x0b\x32/.tecton_proto.auth.ResourceTypePermissionValuesR\x1cresourceTypePermissionValues2\x93\x0c\n\x14\x41uthorizationService\x12]\n\x08GetRoles\x12\".tecton_proto.auth.GetRolesRequest\x1a#.tecton_proto.auth.GetRolesResponse\"\x08\xa2\xbc\xe6\xc0\x05\x02\x10\x01\x12u\n\x10GetAssignedRoles\x12*.tecton_proto.auth.GetAssignedRolesRequest\x1a+.tecton_proto.auth.GetAssignedRolesResponse\"\x08\xa2\xbc\xe6\xc0\x05\x02@\x01\x12x\n\x11ListAssignedRoles\x12+.tecton_proto.auth.ListAssignedRolesRequest\x1a,.tecton_proto.auth.ListAssignedRolesResponse\"\x08\xa2\xbc\xe6\xc0\x05\x02@\x01\x12~\n\x0b\x41ssignRoles\x12%.tecton_proto.auth.AssignRolesRequest\x1a&.tecton_proto.auth.AssignRolesResponse\" \xa2\xbc\xe6\xc0\x05\x02@\x01\xf2\x99\xbd\x46\x13\x08\x01\x12\x0c\x61ssign_roles\x1a\x01\x31\x12\x86\x01\n\rUnassignRoles\x12\'.tecton_proto.auth.UnassignRolesRequest\x1a(.tecton_proto.auth.UnassignRolesResponse\"\"\xa2\xbc\xe6\xc0\x05\x02@\x01\xf2\x99\xbd\x46\x15\x08\x01\x12\x0eunassign_roles\x1a\x01\x31\x12\x8b\x01\n\x0e\x41ssignRolesPut\x12(.tecton_proto.auth.AssignRolesPutRequest\x1a).tecton_proto.auth.AssignRolesPutResponse\"$\xa2\xbc\xe6\xc0\x05\x02@\x01\xf2\x99\xbd\x46\x17\x08\x01\x12\x10\x61ssign_roles_put\x1a\x01\x31\x12\x87\x01\n\x16GetAuthorizedResources\x12\x30.tecton_proto.auth.GetAuthorizedResourcesRequest\x1a\x31.tecton_proto.auth.GetAuthorizedResourcesResponse\"\x08\xa2\xbc\xe6\xc0\x05\x02@\x01\x12\x84\x01\n\x15GetAssignedPrincipals\x12/.tecton_proto.auth.GetAssignedPrincipalsRequest\x1a\x30.tecton_proto.auth.GetAssignedPrincipalsResponse\"\x08\xa2\xbc\xe6\xc0\x05\x02@\x01\x12\x87\x01\n\x16ListAssignedPrincipals\x12\x30.tecton_proto.auth.ListAssignedPrincipalsRequest\x1a\x31.tecton_proto.auth.ListAssignedPrincipalsResponse\"\x08\xa2\xbc\xe6\xc0\x05\x02@\x01\x12x\n\x11GetAppPermissions\x12+.tecton_proto.auth.GetAppPermissionsRequest\x1a,.tecton_proto.auth.GetAppPermissionsResponse\"\x08\xa2\xbc\xe6\xc0\x05\x02@\x01\x12\x8a\x01\n\x17GetWorkspacePermissions\x12\x31.tecton_proto.auth.GetWorkspacePermissionsRequest\x1a\x32.tecton_proto.auth.GetWorkspacePermissionsResponse\"\x08\xa2\xbc\xe6\xc0\x05\x02@\x01\x12r\n\x0fGetIsAuthorized\x12).tecton_proto.auth.GetIsAuthorizedRequest\x1a*.tecton_proto.auth.GetIsAuthorizedResponse\"\x08\xa2\xbc\xe6\xc0\x05\x02@\x01\x42V\n\x0f\x63om.tecton.authB\x19\x41uthorizationServiceProtoP\x01Z&github.com/tecton-ai/tecton_proto/auth')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.auth.authorization_service__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.authB\031AuthorizationServiceProtoP\001Z&github.com/tecton-ai/tecton_proto/auth'
  _ASSIGNROLESREQUEST.fields_by_name['assignments']._options = None
  _ASSIGNROLESREQUEST.fields_by_name['assignments']._serialized_options = b'\352\231\275F\002\010\002'
  _UNASSIGNROLESREQUEST.fields_by_name['assignments']._options = None
  _UNASSIGNROLESREQUEST.fields_by_name['assignments']._serialized_options = b'\352\231\275F\002\010\002'
  _ASSIGNMENT.fields_by_name['resource_type']._options = None
  _ASSIGNMENT.fields_by_name['resource_type']._serialized_options = b'\352\231\275F\002\010\002'
  _ASSIGNMENT.fields_by_name['resource_id']._options = None
  _ASSIGNMENT.fields_by_name['resource_id']._serialized_options = b'\352\231\275F\002\010\002'
  _ASSIGNMENT.fields_by_name['role']._options = None
  _ASSIGNMENT.fields_by_name['role']._serialized_options = b'\352\231\275F\002\010\002\352\231\275F\002\020\001'
  _ASSIGNMENT.fields_by_name['principal_type']._options = None
  _ASSIGNMENT.fields_by_name['principal_type']._serialized_options = b'\352\231\275F\002\010\002'
  _ASSIGNMENT.fields_by_name['principal_id']._options = None
  _ASSIGNMENT.fields_by_name['principal_id']._serialized_options = b'\352\231\275F\002\010\002'
  _ASSIGNROLESPUTREQUEST.fields_by_name['resource_type']._options = None
  _ASSIGNROLESPUTREQUEST.fields_by_name['resource_type']._serialized_options = b'\352\231\275F\002\010\002'
  _ASSIGNROLESPUTREQUEST.fields_by_name['resource_id']._options = None
  _ASSIGNROLESPUTREQUEST.fields_by_name['resource_id']._serialized_options = b'\352\231\275F\002\010\002'
  _ASSIGNROLESPUTREQUEST.fields_by_name['principal_type']._options = None
  _ASSIGNROLESPUTREQUEST.fields_by_name['principal_type']._serialized_options = b'\352\231\275F\002\010\002'
  _ASSIGNROLESPUTREQUEST.fields_by_name['principal_id']._options = None
  _ASSIGNROLESPUTREQUEST.fields_by_name['principal_id']._serialized_options = b'\352\231\275F\002\010\002'
  _ASSIGNROLESPUTREQUEST.fields_by_name['roles']._options = None
  _ASSIGNROLESPUTREQUEST.fields_by_name['roles']._serialized_options = b'\352\231\275F\002\010\002\352\231\275F\002\020\001'
  _AUTHORIZATIONSERVICE.methods_by_name['GetRoles']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['GetRoles']._serialized_options = b'\242\274\346\300\005\002\020\001'
  _AUTHORIZATIONSERVICE.methods_by_name['GetAssignedRoles']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['GetAssignedRoles']._serialized_options = b'\242\274\346\300\005\002@\001'
  _AUTHORIZATIONSERVICE.methods_by_name['ListAssignedRoles']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['ListAssignedRoles']._serialized_options = b'\242\274\346\300\005\002@\001'
  _AUTHORIZATIONSERVICE.methods_by_name['AssignRoles']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['AssignRoles']._serialized_options = b'\242\274\346\300\005\002@\001\362\231\275F\023\010\001\022\014assign_roles\032\0011'
  _AUTHORIZATIONSERVICE.methods_by_name['UnassignRoles']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['UnassignRoles']._serialized_options = b'\242\274\346\300\005\002@\001\362\231\275F\025\010\001\022\016unassign_roles\032\0011'
  _AUTHORIZATIONSERVICE.methods_by_name['AssignRolesPut']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['AssignRolesPut']._serialized_options = b'\242\274\346\300\005\002@\001\362\231\275F\027\010\001\022\020assign_roles_put\032\0011'
  _AUTHORIZATIONSERVICE.methods_by_name['GetAuthorizedResources']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['GetAuthorizedResources']._serialized_options = b'\242\274\346\300\005\002@\001'
  _AUTHORIZATIONSERVICE.methods_by_name['GetAssignedPrincipals']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['GetAssignedPrincipals']._serialized_options = b'\242\274\346\300\005\002@\001'
  _AUTHORIZATIONSERVICE.methods_by_name['ListAssignedPrincipals']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['ListAssignedPrincipals']._serialized_options = b'\242\274\346\300\005\002@\001'
  _AUTHORIZATIONSERVICE.methods_by_name['GetAppPermissions']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['GetAppPermissions']._serialized_options = b'\242\274\346\300\005\002@\001'
  _AUTHORIZATIONSERVICE.methods_by_name['GetWorkspacePermissions']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['GetWorkspacePermissions']._serialized_options = b'\242\274\346\300\005\002@\001'
  _AUTHORIZATIONSERVICE.methods_by_name['GetIsAuthorized']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['GetIsAuthorized']._serialized_options = b'\242\274\346\300\005\002@\001'
  _GETROLESREQUEST._serialized_start=307
  _GETROLESREQUEST._serialized_end=324
  _GETROLESRESPONSE._serialized_start=326
  _GETROLESRESPONSE._serialized_end=401
  _ROLEDEFINITION._serialized_start=404
  _ROLEDEFINITION._serialized_end=793
  _PERMISSIONDEFINITION._serialized_start=795
  _PERMISSIONDEFINITION._serialized_end=904
  _GETASSIGNEDROLESREQUEST._serialized_start=907
  _GETASSIGNEDROLESREQUEST._serialized_end=1167
  _GETASSIGNEDROLESRESPONSE._serialized_start=1169
  _GETASSIGNEDROLESRESPONSE._serialized_end=1276
  _LISTASSIGNEDROLESREQUEST._serialized_start=1279
  _LISTASSIGNEDROLESREQUEST._serialized_end=1540
  _LISTASSIGNEDROLESRESPONSE._serialized_start=1542
  _LISTASSIGNEDROLESRESPONSE._serialized_end=1652
  _GETISAUTHORIZEDREQUEST._serialized_start=1655
  _GETISAUTHORIZEDREQUEST._serialized_end=1852
  _GETISAUTHORIZEDRESPONSE._serialized_start=1854
  _GETISAUTHORIZEDRESPONSE._serialized_end=1944
  _ASSIGNROLESREQUEST._serialized_start=1946
  _ASSIGNROLESREQUEST._serialized_end=2040
  _ASSIGNROLESRESPONSE._serialized_start=2042
  _ASSIGNROLESRESPONSE._serialized_end=2063
  _UNASSIGNROLESREQUEST._serialized_start=2065
  _UNASSIGNROLESREQUEST._serialized_end=2161
  _UNASSIGNROLESRESPONSE._serialized_start=2163
  _UNASSIGNROLESRESPONSE._serialized_end=2186
  _ASSIGNMENT._serialized_start=2189
  _ASSIGNMENT._serialized_end=2484
  _ASSIGNROLESPUTREQUEST._serialized_start=2487
  _ASSIGNROLESPUTREQUEST._serialized_end=2795
  _ASSIGNROLESPUTRESPONSE._serialized_start=2797
  _ASSIGNROLESPUTRESPONSE._serialized_end=2821
  _GETAUTHORIZEDRESOURCESREQUEST._serialized_start=2824
  _GETAUTHORIZEDRESOURCESREQUEST._serialized_end=3057
  _GETAUTHORIZEDRESOURCESRESPONSE._serialized_start=3059
  _GETAUTHORIZEDRESOURCESRESPONSE._serialized_end=3182
  _AUTHORIZEDRESOURCES._serialized_start=3184
  _AUTHORIZEDRESOURCES._serialized_end=3264
  _GETASSIGNEDPRINCIPALSREQUEST._serialized_start=3267
  _GETASSIGNEDPRINCIPALSREQUEST._serialized_end=3497
  _GETASSIGNEDPRINCIPALSRESPONSE._serialized_start=3499
  _GETASSIGNEDPRINCIPALSRESPONSE._serialized_end=3600
  _LISTASSIGNEDPRINCIPALSREQUEST._serialized_start=3603
  _LISTASSIGNEDPRINCIPALSREQUEST._serialized_end=3834
  _LISTASSIGNEDPRINCIPALSRESPONSE._serialized_start=3836
  _LISTASSIGNEDPRINCIPALSRESPONSE._serialized_end=3940
  _ASSIGNMENTBASIC._serialized_start=3943
  _ASSIGNMENTBASIC._serialized_end=4137
  _ASSIGNMENTBASICV2._serialized_start=4140
  _ASSIGNMENTBASICV2._serialized_end=4316
  _GETAPPPERMISSIONSREQUEST._serialized_start=4319
  _GETAPPPERMISSIONSREQUEST._serialized_end=4557
  _RESOURCETYPEPERMISSIONS._serialized_start=4559
  _RESOURCETYPEPERMISSIONS._serialized_end=4680
  _GETAPPPERMISSIONSRESPONSE._serialized_start=4683
  _GETAPPPERMISSIONSRESPONSE._serialized_end=4830
  _RESOURCETYPEPERMISSIONVALUES._serialized_start=4833
  _RESOURCETYPEPERMISSIONVALUES._serialized_end=5014
  _PERMISSION._serialized_start=5017
  _PERMISSION._serialized_end=5156
  _PERMISSIONVALUE._serialized_start=5158
  _PERMISSIONVALUE._serialized_end=5236
  _GETWORKSPACEPERMISSIONSREQUEST._serialized_start=5239
  _GETWORKSPACEPERMISSIONSREQUEST._serialized_end=5518
  _GETWORKSPACEPERMISSIONSRESPONSE._serialized_start=5521
  _GETWORKSPACEPERMISSIONSRESPONSE._serialized_end=5674
  _AUTHORIZATIONSERVICE._serialized_start=5677
  _AUTHORIZATIONSERVICE._serialized_end=7232
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/feature_server/configuration/feature_server_configuration__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.args import pipeline__client_pb2 as tecton__proto_dot_args_dot_pipeline____client__pb2
from tecton_proto.args import transformation__client_pb2 as tecton__proto_dot_args_dot_transformation____client__pb2
from tecton_proto.args import user_defined_function__client_pb2 as tecton__proto_dot_args_dot_user__defined__function____client__pb2
from tecton_proto.auth import acl__client_pb2 as tecton__proto_dot_auth_dot_acl____client__pb2
from tecton_proto.common import aggregation_function__client_pb2 as tecton__proto_dot_common_dot_aggregation__function____client__pb2
from tecton_proto.common import data_type__client_pb2 as tecton__proto_dot_common_dot_data__type____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.common import schema__client_pb2 as tecton__proto_dot_common_dot_schema____client__pb2
from tecton_proto.common import time_window__client_pb2 as tecton__proto_dot_common_dot_time__window____client__pb2
from tecton_proto.data import feature_service__client_pb2 as tecton__proto_dot_data_dot_feature__service____client__pb2
from tecton_proto.data import feature_view__client_pb2 as tecton__proto_dot_data_dot_feature__view____client__pb2
from tecton_proto.data import realtime_compute__client_pb2 as tecton__proto_dot_data_dot_realtime__compute____client__pb2
from tecton_proto.data import tecton_api_key__client_pb2 as tecton__proto_dot_data_dot_tecton__api__key____client__pb2
from tecton_proto.data import transformation__client_pb2 as tecton__proto_dot_data_dot_transformation____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nTtecton_proto/feature_server/configuration/feature_server_configuration__client.proto\x12)tecton_proto.feature_server.configuration\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a(tecton_proto/args/pipeline__client.proto\x1a.tecton_proto/args/transformation__client.proto\x1a\x35tecton_proto/args/user_defined_function__client.proto\x1a#tecton_proto/auth/acl__client.proto\x1a\x36tecton_proto/common/aggregation_function__client.proto\x1a+tecton_proto/common/data_type__client.proto\x1a$tecton_proto/common/id__client.proto\x1a(tecton_proto/common/schema__client.proto\x1a-tecton_proto/common/time_window__client.proto\x1a/tecton_proto/data/feature_service__client.proto\x1a,tecton_proto/data/feature_view__client.proto\x1a\x30tecton_proto/data/realtime_compute__client.proto\x1a.tecton_proto/data/tecton_api_key__client.proto\x1a.tecton_proto/data/transformation__client.proto\"\xea\x01\n\x0c\x46\x65\x61turesPlan\x12[\n\x0c\x66\x65\x61ture_plan\x18\x01 \x01(\x0b\x32\x36.tecton_proto.feature_server.configuration.FeaturePlanH\x00R\x0b\x66\x65\x61turePlan\x12w\n\x16realtime_features_plan\x18\x02 \x01(\x0b\x32?.tecton_proto.feature_server.configuration.RealtimeFeaturesPlanH\x00R\x14realtimeFeaturesPlanB\x04\n\x02\x66p\"\xb0\x02\n\x06\x43olumn\x12:\n\tdata_type\x18\x05 \x01(\x0b\x32\x1d.tecton_proto.common.DataTypeR\x08\x64\x61taType\x12\x35\n\x17\x66\x65\x61ture_view_space_name\x18\x02 \x01(\tR\x14\x66\x65\x61tureViewSpaceName\x12;\n\x1a\x66\x65\x61ture_service_space_name\x18\x03 \x01(\tR\x17\x66\x65\x61tureServiceSpaceName\x12,\n\x12\x66\x65\x61ture_view_index\x18\x04 \x01(\x05R\x10\x66\x65\x61tureViewIndex\x12\x42\n\x1e\x62\x61tch_table_feature_view_index\x18\x06 \x01(\x05R\x1a\x62\x61tchTableFeatureViewIndexJ\x04\x08\x01\x10\x02\"\x92\x11\n\x0b\x46\x65\x61turePlan\x12V\n\routput_column\x18\x04 \x01(\x0b\x32\x31.tecton_proto.feature_server.configuration.ColumnR\x0coutputColumn\x12V\n\rinput_columns\x18\x01 \x03(\x0b\x32\x31.tecton_proto.feature_server.configuration.ColumnR\x0cinputColumns\x12[\n\x14\x61ggregation_function\x18\x03 \x01(\x0e\x32(.tecton_proto.common.AggregationFunctionR\x13\x61ggregationFunction\x12n\n\x1b\x61ggregation_function_params\x18\x06 \x01(\x0b\x32..tecton_proto.common.AggregationFunctionParamsR\x19\x61ggregationFunctionParams\x12H\n\x12\x61ggregation_window\x18\x05 \x01(\x0b\x32\x19.google.protobuf.DurationR\x11\x61ggregationWindow\x12N\n\tjoin_keys\x18\x07 \x03(\x0b\x32\x31.tecton_proto.feature_server.configuration.ColumnR\x08joinKeys\x12_\n\x12wildcard_join_keys\x18\x08 \x03(\x0b\x32\x31.tecton_proto.feature_server.configuration.ColumnR\x10wildcardJoinKeys\x12m\n\x19\x61ggregation_secondary_key\x18\x18 \x01(\x0b\x32\x31.tecton_proto.feature_server.configuration.ColumnR\x17\x61ggregationSecondaryKey\x12\x35\n\x17is_secondary_key_output\x18\x19 \x01(\x08R\x14isSecondaryKeyOutput\x12\x1d\n\ntable_name\x18\t \x01(\tR\ttableName\x12|\n\x19\x64\x61ta_table_timestamp_type\x18\x0f \x01(\x0e\x32\x41.tecton_proto.feature_server.configuration.DataTableTimestampTypeR\x16\x64\x61taTableTimestampType\x12\x82\x01\n\x1bstatus_table_timestamp_type\x18\x10 \x01(\x0e\x32\x43.tecton_proto.feature_server.configuration.StatusTableTimestampTypeR\x18statusTableTimestampType\x12#\n\rtimestamp_key\x18\x0c \x01(\tR\x0ctimestampKey\x12<\n\x0cslide_period\x18\r \x01(\x0b\x32\x19.google.protobuf.DurationR\x0bslidePeriod\x12:\n\x0bserving_ttl\x18\x0e \x01(\x0b\x32\x19.google.protobuf.DurationR\nservingTtl\x12\x30\n\x14refresh_status_table\x18\x11 \x01(\x08R\x12refreshStatusTable\x12*\n\x11\x66\x65\x61ture_view_name\x18\x15 \x01(\tR\x0f\x66\x65\x61tureViewName\x12&\n\x0f\x66\x65\x61ture_view_id\x18\x17 \x01(\tR\rfeatureViewId\x12?\n\x1c\x66\x65\x61ture_store_format_version\x18\x12 \x01(\x05R\x19\x66\x65\x61tureStoreFormatVersion\x12T\n\x13online_store_params\x18\x13 \x01(\x0b\x32$.tecton_proto.data.OnlineStoreParamsR\x11onlineStoreParams\x12.\n\x12\x64\x65letionTimeWindow\x18\x16 \x01(\x03R\x12\x64\x65letionTimeWindow\x12@\n\x0btime_window\x18\x1d \x01(\x0b\x32\x1f.tecton_proto.common.TimeWindowR\ntimeWindow\x12\x64\n\x19\x66\x65\x61ture_view_cache_config\x18\x1b \x01(\x0b\x32).tecton_proto.data.FeatureViewCacheConfigR\x16\x66\x65\x61tureViewCacheConfig\x12\x1f\n\x0b\x63\x61\x63he_index\x18\x1c \x01(\x05R\ncacheIndex\x12\x30\n\x14table_format_version\x18\x1e \x01(\x05R\x12tableFormatVersion\x12(\n\x10\x62\x61tch_table_name\x18\x1f \x01(\tR\x0e\x62\x61tchTableName\x12\x37\n\x18\x62\x61tch_table_window_index\x18  \x01(\x05R\x15\x62\x61tchTableWindowIndex\x12*\n\x11stream_table_name\x18\" \x01(\tR\x0fstreamTableName\x12\x43\n\x05tiles\x18# \x03(\x0b\x32-.tecton_proto.common.OnlineBatchTablePartTileR\x05tiles\x12\x39\n\x19is_compacted_feature_view\x18$ \x01(\x08R\x16isCompactedFeatureView\x12\x35\n\x17\x66\x65\x61ture_set_column_hash\x18! \x01(\tR\x14\x66\x65\x61tureSetColumnHashJ\x04\x08\x02\x10\x03J\x04\x08\n\x10\x0bJ\x04\x08\x0b\x10\x0cJ\x04\x08\x14\x10\x15\"g\n\x11\x46\x65\x61tureVectorPlan\x12R\n\x08\x66\x65\x61tures\x18\x01 \x03(\x0b\x32\x36.tecton_proto.feature_server.configuration.FeaturePlanR\x08\x66\x65\x61tures\"\xbc\x06\n\x14RealtimeFeaturesPlan\x12l\n\x19\x61rgs_from_request_context\x18\x02 \x03(\x0b\x32\x31.tecton_proto.feature_server.configuration.ColumnR\x16\x61rgsFromRequestContext\x12K\n\x07outputs\x18\x03 \x03(\x0b\x32\x31.tecton_proto.feature_server.configuration.ColumnR\x07outputs\x12\x83\x01\n\x12\x66\x65\x61ture_set_inputs\x18\x05 \x03(\x0b\x32U.tecton_proto.feature_server.configuration.RealtimeFeaturesPlan.FeatureSetInputsEntryR\x10\x66\x65\x61tureSetInputs\x12\x37\n\x08pipeline\x18\x06 \x01(\x0b\x32\x1b.tecton_proto.args.PipelineR\x08pipeline\x12K\n\x0ftransformations\x18\x07 \x03(\x0b\x32!.tecton_proto.data.TransformationR\x0ftransformations\x12*\n\x11\x66\x65\x61ture_view_name\x18\x08 \x01(\tR\x0f\x66\x65\x61tureViewName\x12&\n\x0f\x66\x65\x61ture_view_id\x18\t \x01(\tR\rfeatureViewId\x12y\n\x17\x63ompact_transformations\x18\n \x03(\x0b\x32@.tecton_proto.feature_server.configuration.CompactTransformationR\x16\x63ompactTransformations\x1a\x81\x01\n\x15\x46\x65\x61tureSetInputsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12R\n\x05value\x18\x02 \x01(\x0b\x32<.tecton_proto.feature_server.configuration.FeatureVectorPlanR\x05value:\x02\x38\x01J\x04\x08\x01\x10\x02J\x04\x08\x04\x10\x05\"\xee\x01\n\x15\x43ompactTransformation\x12\x44\n\x11transformation_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10transformationId\x12V\n\x13transformation_mode\x18\x02 \x01(\x0e\x32%.tecton_proto.args.TransformationModeR\x12transformationMode\x12\x37\n\x18user_defined_function_id\x18\x03 \x01(\tR\x15userDefinedFunctionId\"\x81\x01\n\rLoggingConfig\x12\x1f\n\x0bsample_rate\x18\x01 \x01(\x02R\nsampleRate\x12.\n\x13log_effective_times\x18\x02 \x01(\x08R\x11logEffectiveTimes\x12\x1f\n\x0b\x61vro_schema\x18\x03 \x01(\tR\navroSchema\"\xd7\x06\n\x12\x46\x65\x61tureServicePlan\x12\x45\n\x12\x66\x65\x61ture_service_id\x18\x04 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10\x66\x65\x61tureServiceId\x12?\n\x0f\x66\x65\x61ture_view_id\x18\n \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12\x30\n\x14\x66\x65\x61ture_service_name\x18\x02 \x01(\tR\x12\x66\x65\x61tureServiceName\x12*\n\x11\x66\x65\x61ture_view_name\x18\t \x01(\tR\x0f\x66\x65\x61tureViewName\x12%\n\x0eworkspace_name\x18\x08 \x01(\tR\rworkspaceName\x12\x45\n\x12workspace_state_id\x18\x11 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10workspaceStateId\x12^\n\x0e\x66\x65\x61tures_plans\x18\r \x03(\x0b\x32\x37.tecton_proto.feature_server.configuration.FeaturesPlanR\rfeaturesPlans\x12N\n\x11join_key_template\x18\x05 \x01(\x0b\x32\".tecton_proto.data.JoinKeyTemplateR\x0fjoinKeyTemplate\x12_\n\x0elogging_config\x18\x0c \x01(\x0b\x32\x38.tecton_proto.feature_server.configuration.LoggingConfigR\rloggingConfig\x12Y\n\x14realtime_environment\x18\x0e \x01(\x0b\x32&.tecton_proto.data.OnlineComputeConfigR\x13realtimeEnvironment\x12\x63\n\x0b\x63\x61\x63he_plans\x18\x10 \x03(\x0b\x32\x42.tecton_proto.feature_server.configuration.FeatureServiceCachePlanR\ncachePlansJ\x04\x08\x01\x10\x02J\x04\x08\x03\x10\x04J\x04\x08\x06\x10\x07J\x04\x08\x07\x10\x08J\x04\x08\x0b\x10\x0c\"\xca\x05\n\x11GlobalTableConfig\x12?\n\x0f\x66\x65\x61ture_view_id\x18\x04 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12*\n\x11\x66\x65\x61ture_view_name\x18\n \x01(\tR\x0f\x66\x65\x61tureViewName\x12%\n\x0eworkspace_name\x18\x0b \x01(\tR\rworkspaceName\x12<\n\x0cslide_period\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0bslidePeriod\x12\x82\x01\n\x1bstatus_table_timestamp_type\x18\x06 \x01(\x0e\x32\x43.tecton_proto.feature_server.configuration.StatusTableTimestampTypeR\x18statusTableTimestampType\x12\x30\n\x14refresh_status_table\x18\x07 \x01(\x08R\x12refreshStatusTable\x12?\n\x1c\x66\x65\x61ture_store_format_version\x18\x08 \x01(\x05R\x19\x66\x65\x61tureStoreFormatVersion\x12T\n\x13online_store_params\x18\t \x01(\x0b\x32$.tecton_proto.data.OnlineStoreParamsR\x11onlineStoreParams\x12\x30\n\x14table_format_version\x18\x0c \x01(\x05R\x12tableFormatVersion\x12Q\n\x17\x66\x65\x61ture_data_water_mark\x18\r \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x14\x66\x65\x61tureDataWaterMarkJ\x04\x08\x01\x10\x02J\x04\x08\x03\x10\x04J\x04\x08\x05\x10\x06\"\x87\x01\n\x12\x46\x65\x61tureServiceAcls\x12\x45\n\x12\x66\x65\x61ture_service_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10\x66\x65\x61tureServiceId\x12*\n\x04\x61\x63ls\x18\x02 \x03(\x0b\x32\x16.tecton_proto.auth.AclR\x04\x61\x63ls\"b\n\rWorkspaceAcls\x12%\n\x0eworkspace_name\x18\x01 \x01(\tR\rworkspaceName\x12*\n\x04\x61\x63ls\x18\x02 \x03(\x0b\x32\x16.tecton_proto.auth.AclR\x04\x61\x63ls\"\xe1\x01\n\x0c\x43\x61naryConfig\x12\x37\n\x18\x66\x65\x61ture_server_canary_id\x18\x01 \x01(\tR\x15\x66\x65\x61tureServerCanaryId\x12\x42\n\x1e\x66\x65\x61ture_server_canary_pod_name\x18\x02 \x01(\tR\x1a\x66\x65\x61tureServerCanaryPodName\x12T\n\'feature_server_canary_follower_endpoint\x18\x03 \x01(\tR#featureServerCanaryFollowerEndpoint\"^\n\x0b\x43\x61\x63heParams\x12;\n\x05redis\x18\x02 \x01(\x0b\x32#.tecton_proto.data.RedisOnlineStoreH\x00R\x05redisB\x0c\n\ncache_typeJ\x04\x08\x01\x10\x02\"\x8c\x0f\n\x1a\x46\x65\x61tureServerConfiguration\x12?\n\rcomputed_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0c\x63omputedTime\x12h\n\x10\x66\x65\x61ture_services\x18\x01 \x03(\x0b\x32=.tecton_proto.feature_server.configuration.FeatureServicePlanR\x0f\x66\x65\x61tureServices\x12\xa0\x01\n\x1bglobal_table_config_by_name\x18\x03 \x03(\x0b\x32\x62.tecton_proto.feature_server.configuration.FeatureServerConfiguration.GlobalTableConfigByNameEntryR\x17globalTableConfigByName\x12O\n\x13\x61uthorized_api_keys\x18\x04 \x03(\x0b\x32\x1f.tecton_proto.data.TectonApiKeyR\x11\x61uthorizedApiKeys\x12o\n\x14\x66\x65\x61ture_service_acls\x18\x05 \x03(\x0b\x32=.tecton_proto.feature_server.configuration.FeatureServiceAclsR\x12\x66\x65\x61tureServiceAcls\x12_\n\x0eworkspace_acls\x18\x06 \x03(\x0b\x32\x38.tecton_proto.feature_server.configuration.WorkspaceAclsR\rworkspaceAcls\x12[\n\x17\x61ll_online_store_params\x18\x07 \x03(\x0b\x32$.tecton_proto.data.OnlineStoreParamsR\x14\x61llOnlineStoreParams\x12x\n\x1c\x66\x65\x61ture_server_canary_config\x18\x08 \x01(\x0b\x32\x37.tecton_proto.feature_server.configuration.CanaryConfigR\x19\x66\x65\x61tureServerCanaryConfig\x12\x64\n\x16remote_compute_configs\x18\t \x03(\x0b\x32..tecton_proto.data.RemoteFunctionComputeConfigR\x14remoteComputeConfigs\x12\x63\n\x1a\x61ll_online_compute_configs\x18\r \x03(\x0b\x32&.tecton_proto.data.OnlineComputeConfigR\x17\x61llOnlineComputeConfigs\x12y\n\x0c\x63\x61\x63he_groups\x18\n \x03(\x0b\x32V.tecton_proto.feature_server.configuration.FeatureServerConfiguration.CacheGroupsEntryR\x0b\x63\x61\x63heGroups\x12\x9c\x01\n\x19user_defined_function_map\x18\x0b \x03(\x0b\x32\x61.tecton_proto.feature_server.configuration.FeatureServerConfiguration.UserDefinedFunctionMapEntryR\x16userDefinedFunctionMap\x12\x63\n\x04jwks\x18\x0c \x03(\x0b\x32O.tecton_proto.feature_server.configuration.FeatureServerConfiguration.JwksEntryR\x04jwks\x1a\x88\x01\n\x1cGlobalTableConfigByNameEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12R\n\x05value\x18\x02 \x01(\x0b\x32<.tecton_proto.feature_server.configuration.GlobalTableConfigR\x05value:\x02\x38\x01\x1au\n\x10\x43\x61\x63heGroupsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12K\n\x05value\x18\x02 \x01(\x0b\x32\x35.tecton_proto.feature_server.configuration.CacheGroupR\x05value:\x02\x38\x01\x1aq\n\x1bUserDefinedFunctionMapEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12<\n\x05value\x18\x02 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x05value:\x02\x38\x01\x1ag\n\tJwksEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x44\n\x05value\x18\x02 \x01(\x0b\x32..tecton_proto.feature_server.configuration.JwkR\x05value:\x02\x38\x01\"7\n\x03Jwk\x12\x0c\n\x01n\x18\x01 \x01(\tR\x01n\x12\x10\n\x03\x61lg\x18\x02 \x01(\tR\x03\x61lg\x12\x10\n\x03use\x18\x03 \x01(\tR\x03use\"/\n\x10RemappedJoinKeys\x12\x1b\n\tjoin_keys\x18\x01 \x03(\tR\x08joinKeys\"\x9c\x02\n\x17\x46\x65\x61tureServiceCachePlan\x12(\n\x10\x66\x65\x61ture_view_ids\x18\x01 \x03(\tR\x0e\x66\x65\x61tureViewIds\x12(\n\x10\x63\x61\x63he_group_name\x18\x02 \x01(\tR\x0e\x63\x61\x63heGroupName\x12r\n\x17remapped_join_key_lists\x18\x03 \x03(\x0b\x32;.tecton_proto.feature_server.configuration.RemappedJoinKeysR\x14remappedJoinKeyLists\x12\x39\n\x19\x66\x65\x61ture_set_column_hashes\x18\x04 \x03(\tR\x16\x66\x65\x61tureSetColumnHashes\"\xd5\x01\n\nCacheGroup\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x1b\n\tjoin_keys\x18\x02 \x03(\tR\x08joinKeys\x12\x32\n\x07key_ttl\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationR\x06keyTtl\x12\x38\n\nkey_jitter\x18\x04 \x01(\x0b\x32\x19.google.protobuf.DurationR\tkeyJitter\x12(\n\x10\x66\x65\x61ture_view_ids\x18\x05 \x03(\tR\x0e\x66\x65\x61tureViewIds*\x90\x01\n\x16\x44\x61taTableTimestampType\x12%\n!DATA_TABLE_TIMESTAMP_TYPE_UNKNOWN\x10\x00\x12&\n\"DATA_TABLE_TIMESTAMP_TYPE_SORT_KEY\x10\x01\x12\'\n#DATA_TABLE_TIMESTAMP_TYPE_ATTRIBUTE\x10\x02*\xc9\x01\n\x18StatusTableTimestampType\x12\'\n#STATUS_TABLE_TIMESTAMP_TYPE_UNKNOWN\x10\x00\x12(\n$STATUS_TABLE_TIMESTAMP_TYPE_SORT_KEY\x10\x01\x12)\n%STATUS_TABLE_TIMESTAMP_TYPE_ATTRIBUTE\x10\x02\x12/\n+STATUS_TABLE_TIMESTAMP_CONTINUOUS_AGGREGATE\x10\x03\x42l\n(com.tecton.feature_service.configurationP\x01Z>github.com/tecton-ai/tecton_proto/feature_server/configuration')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.feature_server.configuration.feature_server_configuration__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n(com.tecton.feature_service.configurationP\001Z>github.com/tecton-ai/tecton_proto/feature_server/configuration'
  _REALTIMEFEATURESPLAN_FEATURESETINPUTSENTRY._options = None
  _REALTIMEFEATURESPLAN_FEATURESETINPUTSENTRY._serialized_options = b'8\001'
  _FEATURESERVERCONFIGURATION_GLOBALTABLECONFIGBYNAMEENTRY._options = None
  _FEATURESERVERCONFIGURATION_GLOBALTABLECONFIGBYNAMEENTRY._serialized_options = b'8\001'
  _FEATURESERVERCONFIGURATION_CACHEGROUPSENTRY._options = None
  _FEATURESERVERCONFIGURATION_CACHEGROUPSENTRY._serialized_options = b'8\001'
  _FEATURESERVERCONFIGURATION_USERDEFINEDFUNCTIONMAPENTRY._options = None
  _FEATURESERVERCONFIGURATION_USERDEFINEDFUNCTIONMAPENTRY._serialized_options = b'8\001'
  _FEATURESERVERCONFIGURATION_JWKSENTRY._options = None
  _FEATURESERVERCONFIGURATION_JWKSENTRY._serialized_options = b'8\001'
  _DATATABLETIMESTAMPTYPE._serialized_start=9579
  _DATATABLETIMESTAMPTYPE._serialized_end=9723
  _STATUSTABLETIMESTAMPTYPE._serialized_start=9726
  _STATUSTABLETIMESTAMPTYPE._serialized_end=9927
  _FEATURESPLAN._serialized_start=848
  _FEATURESPLAN._serialized_end=1082
  _COLUMN._serialized_start=1085
  _COLUMN._serialized_end=1389
  _FEATUREPLAN._serialized_start=1392
  _FEATUREPLAN._serialized_end=3586
  _FEATUREVECTORPLAN._serialized_start=3588
  _FEATUREVECTORPLAN._serialized_end=3691
  _REALTIMEFEATURESPLAN._serialized_start=3694
  _REALTIMEFEATURESPLAN._serialized_end=4522
  _REALTIMEFEATURESPLAN_FEATURESETINPUTSENTRY._serialized_start=4381
  _REALTIMEFEATURESPLAN_FEATURESETINPUTSENTRY._serialized_end=4510
  _COMPACTTRANSFORMATION._serialized_start=4525
  _COMPACTTRANSFORMATION._serialized_end=4763
  _LOGGINGCONFIG._serialized_start=4766
  _LOGGINGCONFIG._serialized_end=4895
  _FEATURESERVICEPLAN._serialized_start=4898
  _FEATURESERVICEPLAN._serialized_end=5753
  _GLOBALTABLECONFIG._serialized_start=5756
  _GLOBALTABLECONFIG._serialized_end=6470
  _FEATURESERVICEACLS._serialized_start=6473
  _FEATURESERVICEACLS._serialized_end=6608
  _WORKSPACEACLS._serialized_start=6610
  _WORKSPACEACLS._serialized_end=6708
  _CANARYCONFIG._serialized_start=6711
  _CANARYCONFIG._serialized_end=6936
  _CACHEPARAMS._serialized_start=6938
  _CACHEPARAMS._serialized_end=7032
  _FEATURESERVERCONFIGURATION._serialized_start=7035
  _FEATURESERVERCONFIGURATION._serialized_end=8967
  _FEATURESERVERCONFIGURATION_GLOBALTABLECONFIGBYNAMEENTRY._serialized_start=8492
  _FEATURESERVERCONFIGURATION_GLOBALTABLECONFIGBYNAMEENTRY._serialized_end=8628
  _FEATURESERVERCONFIGURATION_CACHEGROUPSENTRY._serialized_start=8630
  _FEATURESERVERCONFIGURATION_CACHEGROUPSENTRY._serialized_end=8747
  _FEATURESERVERCONFIGURATION_USERDEFINEDFUNCTIONMAPENTRY._serialized_start=8749
  _FEATURESERVERCONFIGURATION_USERDEFINEDFUNCTIONMAPENTRY._serialized_end=8862
  _FEATURESERVERCONFIGURATION_JWKSENTRY._serialized_start=8864
  _FEATURESERVERCONFIGURATION_JWKSENTRY._serialized_end=8967
  _JWK._serialized_start=8969
  _JWK._serialized_end=9024
  _REMAPPEDJOINKEYS._serialized_start=9026
  _REMAPPEDJOINKEYS._serialized_end=9073
  _FEATURESERVICECACHEPLAN._serialized_start=9076
  _FEATURESERVICECACHEPLAN._serialized_end=9360
  _CACHEGROUP._serialized_start=9363
  _CACHEGROUP._serialized_end=9576
# @@protoc_insertion_point(module_scope)
